# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import re
import os
import glob
import pandas as pd
import hatchet.graphframe
from hatchet.node import Node
from hatchet.graph import Graph
from hatchet.frame import Frame


class TAUReader:
    """Read in a profile generated using TAU."""

    def __init__(self, dirname):
        self.dirname = dirname
        self.node_dicts = []
        self.callpath_to_node = {}
        self.rank_thread_to_data = {}
        self.filepath_to_data = {}
        self.inc_metrics = []
        self.exc_metrics = []
        self.columns = []
        self.multiple_ranks = False
        self.multiple_threads = False

    def create_node_dict(
        self,
        node,
        columns,
        metric_values,
        name,
        filename,
        module,
        start_line,
        end_line,
        rank,
        thread,
    ):
        node_dict = {
            "node": node,
            "rank": rank,
            "thread": thread,
            "name": name,
            "file": filename,
            "module": module,
            "line": int(start_line),
            "end_line": int(end_line),
        }
        for i in range(len(metric_values)):
            node_dict[columns[i + 1]] = metric_values[i]
        return node_dict

    def create_graph(self):
        def _get_name_file_module(is_parent, node_info, symbol):
            """This function gets the name, file and module information
            for a node using the corresponding line in the output file.
            Example line: [UNWIND] <file> [@] <name> [{<file_or_module>} {<line>}]
            There are several line formats in TAU and this function gets
            the node information considering all these formats for which
            examples are given below.
            """
            name, file, module = None, None, None
            # There are several different formats in TAU outputs.
            # There might be file, line, and module information.
            # The following if-else block covers all possible output
            # formats. Example formats are given in comments.
            if symbol == " [@] ":
                # Check if there is a [@] symbol.
                node_info = node_info.split(symbol)
                # We don't need file and module information if it's a parent node.
                if not is_parent:
                    file = node_info[0].split()[1]
                    if "[{" in node_info[1]:
                        # Sometimes we see file and module information inside of [{}]
                        # Example: [UNWIND] <file> [@] <name> [{<file_or_module>} {<line>}]
                        name_and_module = node_info[1].split(" [{")
                        module = name_and_module[1].split()[0].strip("}")
                    else:
                        # Example: [UNWIND] <file> [@] <name> <module>
                        name_and_module = node_info[1].split()
                        module = name_and_module[1]

                    # Check if module is in file.
                    # Assign None to file if it's .so.
                    # Assign None to module if it's .c.
                    if module in file:
                        if ".so" in file:
                            file = None
                        if ".c" in module:
                            module = None
                    name = "[UNWIND] " + name_and_module[0]
                else:
                    # We just need to take name if it is a parent
                    name = "[UNWIND] " + node_info[1].split()[0]
            elif symbol == " C ":
                # Check if there is a C symbol.
                # "C" symbol means it's a C function.
                node_info = node_info.split(symbol)
                name = node_info[0]
                # We don't need file and module information if it's a parent node.
                if not is_parent:
                    if "[{" in node_info[1]:
                        # Example: <name> C [{<file>} {<line>}]
                        node_info = node_info[1].split()
                        file = node_info[0].strip("}[{")
            else:
                if "[{" in node_info:
                    # If there isn't C or [@]
                    # Example: [<type>] <name> [{} {}]
                    node_info = node_info.split(" [{")
                    name = node_info[0]
                    # We don't need file and module information if it's a parent node.
                    if not is_parent:
                        file = node_info[1].split()[0].strip("}{")
                else:
                    # Example 1: [<type>] <name> <module>
                    # Example 2: [<type>] <name>
                    # Example 3: <name>
                    name = node_info
                    node_info = node_info.split()
                    # We need to take module information from the first example.
                    # Another example is "[CONTEXT] .TAU application" which contradicts
                    # with the first example. So we check if there is "\" symbol which
                    # will show the module information in this case.
                    if len(node_info) == 3 and "/" in name:
                        name = node_info[0] + " " + node_info[1]
                        # We don't need file and module information if it's a parent node.
                        if not is_parent:
                            module = node_info[2]
            return [name, file, module]

        def _get_line_numbers(node_info):
            start_line, end_line = 0, 0
            # There should be [{}] symbols if there is line number information.
            if "[{" in node_info:
                tmp_module_or_file_line = (
                    re.search(r"\{.*\}\]", node_info).group(0).split()
                )
                line_numbers = tmp_module_or_file_line[1].strip("}]").replace("{", "")
                start_line = line_numbers
                if "-" in line_numbers:
                    # Sometimes there is "-" between start line and end line
                    # Example: {341,1}-{396,1}
                    line_numbers = line_numbers.split("-")
                    start_line = line_numbers[0].split(",")[0]
                    end_line = line_numbers[1].split(",")[0]
                else:
                    if "," in line_numbers:
                        # Sometimes we don't have "-".
                        # Example: {15,0}
                        start_line = line_numbers.split(",")[0]
                        end_line = line_numbers.split(",")[1]
            return [start_line, end_line]

        def _create_parent(child_node, parent_callpath):
            """In TAU output, sometimes we see a node as a parent
            in the callpath before we see it as a leaf node. In
            this case, we need to create a hatchet node for the parent.

            We can't create a node_dict for the parent because we don't
            know its metric values when we first see it in a callpath.

            Example: a => b => c "<c_metric_values>"
            Here, if we haven't seen 'b' before, we should create it when we
            create 'c'.

            This function recursively creates parent nodes in a callpath
            until it reaches the already existing parent in that callpath.
            """
            parent_node = self.callpath_to_node.get(parent_callpath)

            # Return if arrives to the parent
            # Else create a parent and add parent/child
            if parent_node is not None:
                parent_node.add_child(child_node)
                child_node.add_parent(parent_node)
                return
            else:
                grand_parent_callpath = parent_callpath[:-1]
                parent_info = parent_callpath[-1]
                parent_name = ""

                if " C " in parent_info:
                    parent_name = _get_name_file_module(True, parent_info, " C ")[0]
                elif " [@] " in parent_info:
                    parent_name = _get_name_file_module(True, parent_info, " [@] ")[0]
                else:
                    parent_name = _get_name_file_module(True, parent_info, "")[0]

                parent_node = Node(
                    Frame({"type": "function", "name": parent_name}), None
                )
                self.callpath_to_node[parent_callpath] = parent_node

                parent_node.add_child(child_node)
                child_node.add_parent(parent_node)
                _create_parent(parent_node, grand_parent_callpath)

        def _construct_column_list(first_rank_filenames):
            """This function constructs columns, exc_metrics, and
            inc_metrics using all metric files of a rank. It gets the
            all metric files of a rank as a tuple and only loads the
            second line (metadata) of these files.
            """
            columns = []
            for file_index in range(len(first_rank_filenames)):
                with open(first_rank_filenames[file_index], "r") as f:
                    # Skip the first line: "192 templated_functions_MULTI_TIME"
                    next(f)
                    # No need to check if the metadata is the same for all metric files.
                    metadata = next(f)

                    # Get first three columns from # Name Calls Subrs Excl Incl ProfileCalls #
                    # ProfileCalls is removed since it is is typically set to 0 and not used.
                    # We only do this once since these column names are the same for all files.
                    if file_index == 0:
                        columns.extend(
                            re.match(r"\#\s(.*)\s\#", metadata).group(1).split(" ")[:-3]
                        )

                    # Example metric_name: "PAPI_L2_TCM"
                    # TODO: Decide if Calls and Subrs should be inc or exc metrics
                    metric_name = re.search(r"<value>(.*?)<\/value>", metadata).group(1)
                    if metric_name == "CPU_TIME" or metric_name == "TIME":
                        metric_name = "time"
                    elif metric_name == "Name":
                        metric_name == "name"
                    columns.extend([metric_name, metric_name + " (inc)"])
                    self.exc_metrics.append(metric_name)
                    self.inc_metrics.append(metric_name + " (inc)")
            return columns

        # dirpath -> returns path of a directory, string
        # dirnames -> returns directory names, list
        # files -> returns filenames in a directory, list
        profile_filenames = []
        for dirpath, dirnames, files in os.walk(self.dirname):
            profiles_in_dir = glob.glob(dirpath + "/profile.*")
            if profiles_in_dir:
                # sort input files in each directory in the same order
                profile_filenames.append(sorted(profiles_in_dir))

        # Store all files in a list of tuples.
        # Each tuple stores all the metric files of a rank.
        # We process one rank at a time.
        # Example: [(metric1/profile.x.0.0, metric2/profile.x.0.0), ...]
        profile_filenames = list(zip(*profile_filenames))

        # Get column information from the metric files of a rank.
        self.columns = _construct_column_list(profile_filenames[0])

        list_roots = []
        prev_rank, prev_thread = 0, 0
        # Example filenames_per_rank: (metric1/profile.x.0.0 ...)
        for filenames_per_rank in profile_filenames:
            file_info = filenames_per_rank[0].split(".")
            rank, thread = int(file_info[-3]), int(file_info[-1])
            if not self.multiple_ranks:
                self.multiple_ranks = True if rank != prev_rank else False
            if not self.multiple_threads:
                self.multiple_threads = True if thread != prev_thread else False

            # Load all files represent a different metric for a rank or a thread.
            # If there are 2 metrics, load metric1\profile.x.0.0 and metric2\profile.x.0.0
            file_data = []
            for f_index in range(len(filenames_per_rank)):
                # Store the lines after metadata.
                file_data.append(open(filenames_per_rank[f_index], "r").readlines()[2:])

            # Get the root information from only the first file to compare them
            # with others.
            # Example: ".TAU application" 1 1 272 15755429 0 GROUP="TAU_DEFAULT"
            root_line = re.match(r"\"(.*)\"\s(.*)\sG", file_data[0][0])
            root_name = root_line.group(1).strip(" ")
            # convert it to a tuple to use it as a key in callpath_to_node dictionary
            root_callpath = tuple([root_name])
            root_values = list(map(int, root_line.group(2).split(" ")[:-1]))

            # After first profile.0.0.0, only get Excl and Incl metric values
            # from other files since other columns will be the same.
            # We assume each metric file of a rank has the same root.
            first_file_root_name = re.search(r"\"(.*?)\"", file_data[0][0]).group(1)
            for f_index in range(1, len(file_data)):
                root_name = re.search(r"\"(.*?)\"", file_data[f_index][0]).group(1)
                # Below assert statement throws an error if the roots are not the
                # same for different metric files.
                # TODO: We need to find a solution if this throws an error.
                assert first_file_root_name == root_name, (
                    "Metric files for a rank has different roots.\n"
                    + "File: "
                    + filenames_per_rank[f_index]
                    + "\nLine: 2"
                )
                root_line = re.match(r"\"(.*)\"\s(.*)\sG", file_data[f_index][0])
                root_values.extend(list(map(int, root_line.group(2).split(" ")[2:4])))

            # Check if the root exists in other ranks.
            # Note that we assume the root is the same for all metric files of a rank.
            if root_callpath not in self.callpath_to_node:
                # Create the root node since it doesn't exist
                root_node = Node(Frame({"name": root_name, "type": "function"}), None)

                # Store callpaths to identify nodes
                self.callpath_to_node[root_callpath] = root_node
                list_roots.append(root_node)
            else:
                # Don't create a new node since it is created earlier
                root_node = self.callpath_to_node.get(root_callpath)

            node_dict = self.create_node_dict(
                root_node,
                self.columns,
                root_values,
                root_name,
                None,
                None,
                0,
                0,
                rank,
                thread,
            )
            self.node_dicts.append(node_dict)

            # Start from the line after root.
            # Iterate over only the first metric file of a rank
            # since the lines should be exactly the same across
            # all metric files of a rank.
            # Uses the same "line_index" for other metric files of a rank.
            for line_index in range(1, len(file_data[0])):
                line = file_data[0][line_index]
                metric_values = []
                # We only parse the lines that has "=>" symbol which shows the callpath info.
                # We just skip the other lines.
                if "=>" in line:
                    # Example: ".TAU application  => foo()  => bar()" 31 0 155019 155019 0 GROUP="TAU_SAMPLE|TAU_CALLPATH"
                    callpath_line_regex = re.match(r"\"(.*)\"\s(.*)\sG", line)
                    # callpath: ".TAU application  => foo()  => bar()"
                    callpath = [
                        name.strip(" ")
                        for name in callpath_line_regex.group(1).split("=>")
                    ]

                    # Example leaf_name: StrToInt [{lulesh-util.cc} {13,1}-{29,1}]
                    leaf_name = callpath[-1]
                    callpath = tuple(callpath)
                    parent_callpath = callpath[:-1]
                    # Don't include the value for ProfileCalls.
                    # metric_values: 31 0 155019 155019
                    metric_values = list(
                        map(float, callpath_line_regex.group(2).split(" ")[:-1])
                    )

                    # Get start and end line information
                    leaf_line_numbers = _get_line_numbers(leaf_name)
                    # Get name, file, and module information using the leaf name
                    # and the symbol on it
                    if " C " in leaf_name:
                        leaf_name_file_module = _get_name_file_module(
                            False, leaf_name, " C "
                        )
                    elif " [@] " in leaf_name:
                        leaf_name_file_module = _get_name_file_module(
                            False, leaf_name, " [@] "
                        )
                    else:
                        leaf_name_file_module = _get_name_file_module(
                            False, leaf_name, ""
                        )

                    # Example: ".TAU application  => foo()  => bar()" 31 0 155019..."
                    first_file_callpath_line = re.search(
                        r"\"(.*?)\"", file_data[0][line_index]
                    ).group(1)
                    # After first profile.x.0.0, only get Excl and Incl metric values
                    # from other files.
                    for f_index in range(1, len(file_data)):
                        other_file_callpath_line = re.search(
                            r"\"(.*?)\"", file_data[f_index][line_index]
                        ).group(1)
                        # We assume metric files of a rank should have the exact same lines.
                        # Only difference should be the Incl and Excl metric values.
                        # TODO: We should find a solution if this raises an error.
                        assert first_file_callpath_line == other_file_callpath_line, (
                            "Lines across metric files for a rank are not the same.\n"
                            + "File: "
                            + filenames_per_rank[f_index]
                            + "\nLine: "
                            + str(line_index + 3)
                        )
                        # Get the information from the same line in each file. "line_index".
                        callpath_line_regex = re.match(
                            r"\"(.*)\"\s(.*)\sG", file_data[f_index][line_index]
                        )
                        metric_values.extend(
                            map(float, callpath_line_regex.group(2).split(" ")[2:4])
                        )

                    leaf_node = self.callpath_to_node.get(callpath)
                    # Check if that node is created earlier
                    if leaf_node is None:
                        # Create the node since it doesn't exist
                        leaf_node = Node(
                            Frame(
                                {"type": "function", "name": leaf_name_file_module[0]}
                            ),
                            None,
                        )
                        self.callpath_to_node[callpath] = leaf_node

                        # Get its parent from its callpath.
                        parent_node = self.callpath_to_node.get(parent_callpath)
                        if parent_node is None:
                            # Create parent if it doesn't exist.
                            _create_parent(leaf_node, parent_callpath)
                        else:
                            parent_node.add_child(leaf_node)
                            leaf_node.add_parent(parent_node)

                    node_dict = self.create_node_dict(
                        leaf_node,
                        self.columns,
                        metric_values,
                        # name
                        leaf_name_file_module[0],
                        # file
                        leaf_name_file_module[1],
                        # module
                        leaf_name_file_module[2],
                        # start line
                        leaf_line_numbers[0],
                        # end line
                        leaf_line_numbers[1],
                        rank,
                        thread,
                    )

                    self.node_dicts.append(node_dict)

        return list_roots

    def read(self):
        """Read the TAU profile file to extract the calling context tree."""
        # Add all nodes and roots.
        roots = self.create_graph()
        # Create a graph object once all nodes have been added.
        graph = Graph(roots)
        graph.enumerate_traverse()

        dataframe = pd.DataFrame.from_dict(data=self.node_dicts)

        indices = []
        # Set indices according to rank/thread numbers.
        if self.multiple_ranks and self.multiple_threads:
            indices = ["node", "rank", "thread"]
        elif self.multiple_ranks:
            dataframe.drop(columns=["thread"], inplace=True)
            indices = ["node", "rank"]
        elif self.multiple_threads:
            dataframe.drop(columns=["rank"], inplace=True)
            indices = ["node", "thread"]
        else:
            indices = ["node"]

        dataframe.set_index(indices, inplace=True)
        dataframe.sort_index(inplace=True)

        # Fill the missing ranks
        # After unstacking and iterating over rows, there
        # will be "NaN" values for some ranks. Find the first
        # rank that has notna value and use it for other rows/ranks
        # of the multiindex.
        # TODO: iterrows() is not the best way to iterate over rows.
        if self.multiple_ranks or self.multiple_threads:
            dataframe = dataframe.unstack()
            for idx, row in dataframe.iterrows():
                # There is always a valid name for an index.
                # Take that valid name and assign to other ranks/rows.
                name = row["name"][row["name"].first_valid_index()]
                dataframe.loc[idx, "name"] = name

                # Sometimes there is no file information.
                if row["file"].first_valid_index() is not None:
                    file = row["file"][row["file"].first_valid_index()]
                    dataframe.loc[idx, "file"] = file

                # Sometimes there is no module information.
                if row["module"].first_valid_index() is not None:
                    module = row["module"][row["module"].first_valid_index()]
                    dataframe.loc[idx, "module"] = module

                # Fill the rest with 0
                dataframe.fillna(0, inplace=True)

            # Stack the dataframe
            dataframe = dataframe.stack()

        default_metric = "time (inc)"

        return hatchet.graphframe.GraphFrame(
            graph, dataframe, self.exc_metrics, self.inc_metrics, default_metric
        )



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import copy
import sys
import traceback

from collections import defaultdict

import pandas as pd
import numpy as np
import multiprocess as mp
import json

from .node import Node
from .graph import Graph
from .frame import Frame
from .query import (
    is_hatchet_query,
    ObjectQuery,
    parse_string_dialect,
    QueryEngine,
    AbstractQuery,
)
from .external.console import ConsoleRenderer
from .util.dot import trees_to_dot
from .util.deprecated import deprecated_params

try:
    from .cython_modules.libs import graphframe_modules as _gfm_cy
except ImportError:
    print("-" * 80)
    print(
        """Error: Shared object (.so) not found for cython module.\n\tPlease run install.sh from the hatchet root directory to build modules."""
    )
    print("-" * 80)
    traceback.print_exc()
    raise


def parallel_apply(filter_function, dataframe, queue):
    """A function called in parallel, which does a pandas apply on part of a
    dataframe and returns the results via multiprocessing queue function."""
    filtered_rows = dataframe.apply(filter_function, axis=1)
    filtered_df = dataframe[filtered_rows]
    queue.put(filtered_df)


class GraphFrame:
    """An input dataset is read into an object of this type, which includes a graph
    and a dataframe.
    """

    def __init__(
        self,
        graph,
        dataframe,
        exc_metrics=None,
        inc_metrics=None,
        default_metric="time",
        metadata={},
    ):
        """Create a new GraphFrame from a graph and a dataframe.

        Likely, you do not want to use this function.

        See ``from_hpctoolkit``, ``from_caliper``, ``from_gprof_dot``, and
        other reader methods for easier ways to create a ``GraphFrame``.

        Arguments:
             graph (Graph): Graph of nodes in this GraphFrame.
             dataframe (DataFrame): Pandas DataFrame indexed by Nodes
                 from the graph, and potentially other indexes.
             exc_metrics: list of names of exclusive metrics in the dataframe.
             inc_metrics: list of names of inclusive metrics in the dataframe.
        """
        if graph is None:
            raise ValueError("GraphFrame() requires a Graph")
        if dataframe is None:
            raise ValueError("GraphFrame() requires a DataFrame")

        if "node" not in list(dataframe.index.names):
            raise ValueError(
                "DataFrames passed to GraphFrame() must have an index called 'node'."
            )

        self.graph = graph
        self.dataframe = dataframe
        self.exc_metrics = [] if exc_metrics is None else exc_metrics
        self.inc_metrics = [] if inc_metrics is None else inc_metrics
        self.default_metric = default_metric
        self.metadata = metadata
        self.query_engine = QueryEngine()

    @staticmethod
    def from_hpctoolkit(dirname):
        """Read an HPCToolkit database directory into a new GraphFrame.

        Arguments:
            dirname (str): parent directory of an HPCToolkit
                experiment.xml file

        Returns:
            (GraphFrame): new GraphFrame containing HPCToolkit profile data
        """
        # import this lazily to avoid circular dependencies
        from .readers.hpctoolkit_reader import HPCToolkitReader

        return HPCToolkitReader(dirname).read()

    @staticmethod
    def from_caliper(filename_or_stream, query=None):
        """Read in a Caliper .cali or .json file.

        Args:
            filename_or_stream (str or file-like): name of a Caliper output
                file in `.cali` or JSON-split format, or an open file object
                to read one
            query (str): cali-query in CalQL format
        """
        # import this lazily to avoid circular dependencies
        from .readers.caliper_reader import CaliperReader

        return CaliperReader(filename_or_stream, query).read()

    @staticmethod
    def from_caliperreader(
        filename_or_caliperreader, native=False, string_attributes=[]
    ):
        """Read in a native Caliper `cali` file using Caliper's python reader.

        Args:
            filename_or_caliperreader (str or CaliperReader): name of a Caliper
                output file in `.cali` format, or a CaliperReader object
            native (bool): use native or user-readable metric names (default)
            string_attributes (str or list, optional): Adds existing string
                attributes from within the caliper file to the dataframe
        """
        # import this lazily to avoid circular dependencies
        from .readers.caliper_native_reader import CaliperNativeReader

        return CaliperNativeReader(
            filename_or_caliperreader, native, string_attributes
        ).read()

    @staticmethod
    def from_spotdb(db_key, list_of_ids=None):
        """Read multiple graph frames from a SpotDB instance

        Args:
            db_key (str or SpotDB object): locator for SpotDB instance
                This can be a SpotDB object directly, or a locator for a spot
                database, which is a string with either:

                    * A directory for .cali files,
                    * A .sqlite file name
                    * A SQL database URL (e.g., "mysql://hostname/db")

            list_of_ids: The list of run IDs to read from the database.
                If this is None, returns all runs.

        Returns:
            A list of graphframes, one for each requested run that was found
        """

        from .readers.spotdb_reader import SpotDBReader

        return SpotDBReader(db_key, list_of_ids).read()

    @staticmethod
    def from_gprof_dot(filename):
        """Read in a DOT file generated by gprof2dot."""
        # import this lazily to avoid circular dependencies
        from .readers.gprof_dot_reader import GprofDotReader

        return GprofDotReader(filename).read()

    @staticmethod
    def from_cprofile(filename):
        """Read in a pstats/prof file generated using python's cProfile."""
        # import this lazily to avoid circular dependencies
        from .readers.cprofile_reader import CProfileReader

        return CProfileReader(filename).read()

    @staticmethod
    def from_pyinstrument(filename):
        """Read in a JSON file generated using Pyinstrument."""
        # import this lazily to avoid circular dependencies
        from .readers.pyinstrument_reader import PyinstrumentReader

        return PyinstrumentReader(filename).read()

    @staticmethod
    def from_tau(dirname):
        """Read in a profile generated using TAU."""
        # import this lazily to avoid circular dependencies
        from .readers.tau_reader import TAUReader

        return TAUReader(dirname).read()

    @staticmethod
    def from_timemory(input=None, select=None, **_kwargs):
        """Read in timemory data.

        Links:
            https://github.com/NERSC/timemory
            https://timemory.readthedocs.io

        Arguments:
            input (str or file-stream or dict or None):
                Valid argument types are:

                1. Filename for a timemory JSON tree file
                2. Open file stream to one of these files
                3. Dictionary from timemory JSON tree


                Currently, timemory supports two JSON layouts: flat and tree.
                The former is a 1D-array representation of the hierarchy which
                represents the hierarchy via indentation schemes in the labels
                and is not compatible with hatchet. The latter is a hierarchical
                representation of the data and is the required JSON layout when
                using hatchet. Timemory JSON tree files typically have the
                extension ".tree.json".

                If input is None, this assumes that timemory has been recording
                data within the application that is using hatchet. In this
                situation, this method will attempt to import the data directly
                from timemory.

                At the time of this writing, the direct data import will:

                1. Stop any currently collecting components
                2. Aggregate child thread data of the calling thread
                3. Clear all data on the child threads
                4. Aggregate the data from any MPI and/or UPC++ ranks.


                Thus, if MPI or UPC++ is used, every rank must call this routine.
                The zeroth rank will have the aggregation and all the other
                non-zero ranks will only have the rank-specific data.

                Whether or not the per-thread and per-rank data itself is
                combined is controlled by the `collapse_threads` and
                `collapse_processes` attributes in the `timemory.settings`
                submodule.

                In the C++ API, it is possible for only #1 to be applied and data
                can be obtained for an individual thread and/or rank without
                aggregation. This is not currently available to Python, however,
                it can be made available upon request via a GitHub Issue.

            select (list of str):
                A list of strings which match the component enumeration names, e.g. ["cpu_clock"].

            per_thread (boolean):
                Ensures that when applying filters to the graphframe, frames with
                identical name/file/line/etc. info but from different threads are
                not combined

            per_rank (boolean):
                Ensures that when applying filters to the graphframe, frames with
                identical name/file/line/etc. info but from different ranks are
                not combined

        """
        from .readers.timemory_reader import TimemoryReader

        if input is not None:
            try:
                return TimemoryReader(input, select, **_kwargs).read()
            except IOError:
                pass
        else:
            try:
                import timemory

                TimemoryReader(timemory.get(hierarchy=True), select, **_kwargs).read()
            except ImportError:
                print(
                    "Error! timemory could not be imported. Provide filename, file stream, or dict."
                )
                raise

    @staticmethod
    def from_literal(graph_dict):
        """Create a GraphFrame from a list of dictionaries."""
        # import this lazily to avoid circular dependencies
        from .readers.literal_reader import LiteralReader

        return LiteralReader(graph_dict).read()

    @staticmethod
    def from_lists(*lists):
        """Make a simple GraphFrame from lists.

        This creates a Graph from lists (see ``Graph.from_lists()``) and uses
        it as the index for a new GraphFrame. Every node in the new graph has
        exclusive time of 1 and inclusive time is computed automatically.

        """
        graph = Graph.from_lists(*lists)
        graph.enumerate_traverse()

        df = pd.DataFrame({"node": list(graph.traverse())})
        df["time"] = [1.0] * len(graph)
        df["name"] = [n.frame["name"] for n in graph.traverse()]
        df.set_index(["node"], inplace=True)
        df.sort_index(inplace=True)

        gf = GraphFrame(graph, df, ["time"], [])
        gf.update_inclusive_columns()
        return gf

    @staticmethod
    def from_json(json_spec, **kwargs):
        from .readers.json_reader import JsonReader

        return JsonReader(json_spec).read(**kwargs)

    @staticmethod
    def from_hdf(filename, **kwargs):
        # import this lazily to avoid circular dependencies
        from .readers.hdf5_reader import HDF5Reader

        return HDF5Reader(filename).read(**kwargs)

    def to_hdf(self, filename, key="hatchet_graphframe", **kwargs):
        # import this lazily to avoid circular dependencies
        from .writers.hdf5_writer import HDF5Writer

        HDF5Writer(filename).write(self, key=key, **kwargs)

    def copy(self):
        """Return a partially shallow copy of the graphframe.

        This copies the DataFrame object, but the data is comprised of references. The Graph is shared between self and the new GraphFrame.

        Arguments:
            self (GraphFrame): Object to make a copy of.

        Returns:
            other (GraphFrame): Copy of self
                graph (graph): Reference to self's graph
                dataframe (DataFrame): Pandas "non-deep" copy of dataframe
                exc_metrics (list): Copy of self's exc_metrics
                inc_metrics (list): Copy of self's inc_metrics
                default_metric (str): N/A
                metadata (dict): Copy of self's metadata
        """
        return GraphFrame(
            self.graph,
            self.dataframe.copy(deep=False),
            copy.copy(self.exc_metrics),
            copy.copy(self.inc_metrics),
            self.default_metric,
            copy.copy(self.metadata),
        )

    def deepcopy(self):
        """Return a deep copy of the graphframe.

        Arguments:
            self (GraphFrame): Object to make a copy of.

        Returns:
            other (GraphFrame): Copy of self
                graph (graph): Deep copy of self's graph
                dataframe (DataFrame): Pandas "deep" copy with node objects updated to match graph from "node_clone"
                exc_metrics (list): Copy of self's exc_metrics
                inc_metrics (list): Copy of self's inc_metrics
                default_metric (str): N/A
                metadata (dict): Copy of self's metadata
        """
        node_clone = {}
        graph_copy = self.graph.copy(node_clone)
        dataframe_copy = self.dataframe.copy()

        index_names = dataframe_copy.index.names
        dataframe_copy.reset_index(inplace=True)

        dataframe_copy["node"] = dataframe_copy["node"].apply(lambda x: node_clone[x])

        dataframe_copy.set_index(index_names, inplace=True)

        return GraphFrame(
            graph_copy,
            dataframe_copy,
            copy.deepcopy(self.exc_metrics),
            copy.deepcopy(self.inc_metrics),
            self.default_metric,
            copy.deepcopy(self.metadata),
        )

    def drop_index_levels(self, function=np.mean):
        """Drop all index levels but `node`."""
        index_names = list(self.dataframe.index.names)
        index_names.remove("node")

        # create dict that stores aggregation function for each column
        agg_dict = {}
        for col in self.dataframe.columns.tolist():
            if col in self.exc_metrics + self.inc_metrics:
                agg_dict[col] = function
            else:
                agg_dict[col] = lambda x: x.iloc[0]

        # perform a groupby to merge nodes that just differ in index columns
        self.dataframe.reset_index(level="node", inplace=True)
        agg_df = self.dataframe.groupby("node").agg(agg_dict)

        self.dataframe = agg_df

    def filter(
        self,
        filter_obj,
        squash=True,
        update_inc_cols=True,
        num_procs=mp.cpu_count(),
        rec_limit=1000,
        multi_index_mode="off",
    ):
        """Filter the dataframe using a user-supplied function.

        Note: Operates in parallel on user-supplied lambda functions.

        Arguments:
            filter_obj (callable, list, or QueryMatcher): the filter to apply to the GraphFrame.
            squash (boolean, optional): if True, automatically call squash for the user.
            update_inc_cols (boolean, optional): if True, update inclusive columns when performing squash.
            rec_limit: set Python recursion limit, increase if running into
                recursion depth errors) (default: 1000).
        """
        sys.setrecursionlimit(rec_limit)

        dataframe_copy = self.dataframe.copy()

        index_names = self.dataframe.index.names
        dataframe_copy.reset_index(inplace=True)

        filtered_df = None

        if callable(filter_obj):
            # applying pandas filter using the callable function
            if num_procs > 1:
                # perform filter in parallel (default)
                queue = mp.Queue()
                processes = []
                returned_frames = []
                subframes = np.array_split(dataframe_copy, num_procs)

                # Manually create a number of processes equal to the number of
                # logical cpus available
                for pid in range(num_procs):
                    process = mp.Process(
                        target=parallel_apply,
                        args=(filter_obj, subframes[pid], queue),
                    )
                    process.start()
                    processes.append(process)

                # Stores filtered subframes in a list: 'returned_frames', for
                # pandas concatenation. This intermediary list is used because
                # pandas concat is faster when called only once on a list of
                # dataframes, than when called multiple times appending onto a
                # frame of increasing size.
                for pid in range(num_procs):
                    returned_frames.append(queue.get())

                for proc in processes:
                    proc.join()

                filtered_df = pd.concat(returned_frames)

            else:
                # perform filter sequentiually if num_procs = 1
                filtered_rows = dataframe_copy.apply(filter_obj, axis=1)
                filtered_df = dataframe_copy[filtered_rows]

        elif isinstance(filter_obj, (list, str)) or is_hatchet_query(filter_obj):
            # use a callpath query to apply the filter
            query = filter_obj
            # If a raw Object-dialect query is provided (not already passed to ObjectQuery),
            # create a new ObjectQuery object.
            if isinstance(filter_obj, list):
                query = ObjectQuery(filter_obj, multi_index_mode)
            # If a raw String-dialect query is provided (not already passed to StringQuery),
            # create a new StringQuery object.
            elif isinstance(filter_obj, str):
                query = parse_string_dialect(filter_obj, multi_index_mode)
            # If an old-style query is provided, extract the underlying new-style query.
            elif issubclass(type(filter_obj), AbstractQuery):
                query = filter_obj._get_new_query()
            query_matches = self.query_engine.apply(query, self.graph, self.dataframe)
            # match_set = list(set().union(*query_matches))
            # filtered_df = dataframe_copy.loc[dataframe_copy["node"].isin(match_set)]
            filtered_df = dataframe_copy.loc[dataframe_copy["node"].isin(query_matches)]
        else:
            raise InvalidFilter(
                "The argument passed to filter must be a callable, a query path list, or a QueryMatcher object."
            )

        if filtered_df.shape[0] == 0:
            raise EmptyFilter(
                "The provided filter would have produced an empty GraphFrame."
            )

        filtered_df.set_index(index_names, inplace=True)

        filtered_gf = GraphFrame(self.graph, filtered_df)
        filtered_gf.exc_metrics = self.exc_metrics
        filtered_gf.inc_metrics = self.inc_metrics
        filtered_gf.default_metric = self.default_metric
        filtered_gf.metadata = self.metadata

        if squash:
            return filtered_gf.squash(update_inc_cols)
        return filtered_gf

    def squash(self, update_inc_cols=True):
        """Rewrite the Graph to include only nodes present in the DataFrame's rows.

        This can be used to simplify the Graph, or to normalize Graph
        indexes between two GraphFrames.

        Arguments:
            update_inc_cols (boolean, optional): if True, update inclusive columns.
        """
        index_names = self.dataframe.index.names
        self.dataframe.reset_index(inplace=True)

        # create new nodes for each unique node in the old dataframe
        old_to_new = {n: n.copy() for n in set(self.dataframe["node"])}
        for i in old_to_new:
            old_to_new[i]._hatchet_nid = i._hatchet_nid

        # Maintain sets of connections to make for each old node.
        # Start with old -> new mapping and update as we traverse subgraphs.
        connections = defaultdict(lambda: set())
        connections.update({k: {v} for k, v in old_to_new.items()})

        new_roots = []  # list of new roots

        # connect new nodes to children according to transitive
        # relationships in the old graph.
        def rewire(node, new_parent, visited):
            # make all transitive connections for the node we're visiting
            for n in connections[node]:
                if new_parent:
                    # there is a parent in the new graph; connect it
                    if n not in new_parent.children:
                        new_parent.add_child(n)
                        n.add_parent(new_parent)

                elif n not in new_roots:
                    # this is a new root
                    new_roots.append(n)

            new_node = old_to_new.get(node)
            transitive = set()
            if node not in visited:
                visited.add(node)
                for child in node.children:
                    transitive |= rewire(child, new_node or new_parent, visited)

            if new_node:
                # since new_node exists in the squashed graph, we only
                # need to connect new_node
                return {new_node}
            else:
                # connect parents to the first transitively reachable
                # new_nodes of nodes we're removing with this squash
                connections[node] |= transitive
                return connections[node]

        # run rewire for each root and make a new graph
        visited = set()
        for root in self.graph.roots:
            rewire(root, None, visited)
        graph = Graph(new_roots)
        graph.enumerate_traverse()

        # reindex new dataframe with new nodes
        df = self.dataframe.copy()
        df["node"] = df["node"].apply(lambda x: old_to_new[x])

        # at this point, the graph is potentially invalid, as some nodes
        # may have children with identical frames.
        merges = graph.normalize()
        df["node"] = df["node"].apply(lambda n: merges.get(n, n))

        self.dataframe.set_index(index_names, inplace=True)
        df.set_index(index_names, inplace=True)
        # create dict that stores aggregation function for each column
        agg_dict = {}
        for col in df.columns.tolist():
            if col in self.exc_metrics + self.inc_metrics:
                # use min_count=1 (default is 0) here, so sum of an all-NA
                # series is NaN, not 0
                # when min_count=1, sum([NaN, NaN)] = NaN
                # when min_count=0, sum([NaN, NaN)] = 0
                agg_dict[col] = lambda x: x.sum(min_count=1)
            else:
                agg_dict[col] = lambda x: x.iloc[0]

        # perform a groupby to merge nodes with the same callpath
        agg_df = df.groupby(index_names).agg(agg_dict)
        agg_df.sort_index(inplace=True)

        # put it all together
        new_gf = GraphFrame(
            graph,
            agg_df,
            self.exc_metrics,
            self.inc_metrics,
            self.default_metric,
            self.metadata,
        )
        if update_inc_cols:
            new_gf.update_inclusive_columns()
        return new_gf

    def _init_sum_columns(self, columns, out_columns):
        """Helper function for subtree_sum and subgraph_sum."""
        if out_columns is None:
            out_columns = columns
        else:
            # init out columns with input columns in case they are not there.
            for col, out in zip(columns, out_columns):
                self.dataframe[out] = self.dataframe[col]

        if len(columns) != len(out_columns):
            raise ValueError("columns out_columns must be the same length!")

        return out_columns

    def subtree_sum(
        self, columns, out_columns=None, function=lambda x: x.sum(min_count=1)
    ):
        """Compute sum of elements in subtrees.  Valid only for trees.

        For each row in the graph, ``out_columns`` will contain the
        element-wise sum of all values in ``columns`` for that row's node
        and all of its descendants.

        This algorithm will multiply count nodes with in-degree higher
        than one -- i.e., it is only correct for trees.  Prefer using
        ``subgraph_sum`` (which calls ``subtree_sum`` if it can), unless
        you have a good reason not to.

        Arguments:
            columns (list of str): names of columns to sum (default: all columns)
            out_columns (list of str): names of columns to store results
                (default: in place)
            function (callable): associative operator used to sum
                elements, sum of an all-NA series is NaN (default: sum(min_count=1))
        """
        out_columns = self._init_sum_columns(columns, out_columns)

        # sum over the output columns
        for node in self.graph.traverse(order="post"):
            if node.children:
                # TODO: need a better way of aggregating inclusive metrics when
                # TODO: there is a multi-index
                try:
                    is_multi_index = isinstance(
                        self.dataframe.index, pd.core.index.MultiIndex
                    )
                except AttributeError:
                    is_multi_index = isinstance(self.dataframe.index, pd.MultiIndex)

                if is_multi_index:
                    for rank_thread in self.dataframe.loc[
                        (node), out_columns
                    ].index.unique():
                        # rank_thread is either rank or a tuple of (rank, thread).
                        # We check if rank_thread is a tuple and if it is, we
                        # create a tuple of (node, rank, thread). If not, we create
                        # a tuple of (node, rank).
                        if isinstance(rank_thread, tuple):
                            df_index1 = (node,) + rank_thread
                            df_index2 = ([node] + node.children,) + rank_thread
                        else:
                            df_index1 = (node, rank_thread)
                            df_index2 = ([node] + node.children, rank_thread)

                        for col in out_columns:
                            self.dataframe.loc[df_index1, col] = function(
                                self.dataframe.loc[df_index2, col]
                            )
                else:
                    for col in out_columns:
                        self.dataframe.loc[node, col] = function(
                            self.dataframe.loc[[node] + node.children, col]
                        )

    def subgraph_sum(
        self, columns, out_columns=None, function=lambda x: x.sum(min_count=1)
    ):
        """Compute sum of elements in subgraphs.

        For each row in the graph, ``out_columns`` will contain the
        element-wise sum of all values in ``columns`` for that row's node
        and all of its descendants.

        This algorithm is worst-case quadratic in the size of the graph,
        so we try to call ``subtree_sum`` if we can.  In general, there
        is not a particularly efficient algorithm known for subgraph
        sums, so this does about as well as we know how.

        Arguments:
            columns (list of str):  names of columns to sum (default: all columns)
            out_columns (list of str): names of columns to store results
                (default: in place)
            function (callable): associative operator used to sum
                elements, sum of an all-NA series is NaN (default: sum(min_count=1))
        """
        if self.graph.is_tree():
            self.subtree_sum(columns, out_columns, function)
            return

        out_columns = self._init_sum_columns(columns, out_columns)
        for node in self.graph.traverse():
            subgraph_nodes = list(node.traverse())
            # TODO: need a better way of aggregating inclusive metrics when
            # TODO: there is a multi-index
            try:
                is_multi_index = isinstance(
                    self.dataframe.index, pd.core.index.MultiIndex
                )
            except AttributeError:
                is_multi_index = isinstance(self.dataframe.index, pd.MultiIndex)

            if is_multi_index:
                for rank_thread in self.dataframe.loc[
                    (node), out_columns
                ].index.unique():
                    # rank_thread is either rank or a tuple of (rank, thread).
                    # We check if rank_thread is a tuple and if it is, we
                    # create a tuple of (node, rank, thread). If not, we create
                    # a tuple of (node, rank).
                    if isinstance(rank_thread, tuple):
                        df_index1 = (node,) + rank_thread
                        df_index2 = (subgraph_nodes,) + rank_thread
                    else:
                        df_index1 = (node, rank_thread)
                        df_index2 = (subgraph_nodes, rank_thread)

                    for col in out_columns:
                        self.dataframe.loc[df_index1, col] = [
                            function(self.dataframe.loc[df_index2, col])
                        ]
            else:
                # TODO: if you take the list constructor away from the
                # TODO: assignment below, this assignment gives NaNs. Why?
                self.dataframe.loc[(node), out_columns] = list(
                    function(self.dataframe.loc[(subgraph_nodes), columns])
                )

    def generate_exclusive_columns(self):
        """Generates exclusive metrics from available inclusive metrics.

        Currently, this function determines which metrics to generate by looking for one of two things:

        1. An inclusive metric ending in "(inc)" that does not have an exclusive metric with the same name (minus "(inc)")
        2. An inclusive metric not ending in "(inc)"

        The metrics that are generated will have one of two name formats:

        1. If the corresponding inclusive metric's name ends in "(inc)", the exclusive metric will have the same
           name, minus "(inc)"
        2. If the corresponding inclusive metric's name does not end in "(inc)", the exclusive metric will have the same
           name as the inclusive metric, followed by a "(exc)" suffix
        """
        # TODO Change how exclusive-inclusive pairs are determined when inc_metrics and exc_metrics are changed
        # Iterate over inclusive metrics and collect tuples of (new exclusive metrics name, inclusive metric name)
        generation_pairs = []
        for inc in self.inc_metrics:
            # If the metric isn't numeric, it is really categorical. This means the inclusive/exclusive thing doesn't really apply.
            if not pd.api.types.is_numeric_dtype(self.dataframe[inc]):
                continue
            # Assume that metrics ending in "(inc)" are generated
            if inc.endswith("(inc)"):
                possible_exc = inc[: -len("(inc)")].strip()
                # If a metric with the same name as the inclusive metrics minus the "(inc)" does not exist in exc_metrics,
                # assume that there is not a corresponding exclusive metric. So, add this new exclusive metric to the generation list.
                if possible_exc not in self.exc_metrics:
                    generation_pairs.append((possible_exc, inc))
            # If there is an inclusive metric without the "(inc)" suffix,
            # assume that there is no corresponding exclusive metric. So, add this new exclusive metrics (with the "(exc)"
            # suffix) to the generation list.
            else:
                generation_pairs.append((inc + " (exc)", inc))
        # Consider each new exclusive metric and its corresponding inclusive metric
        for exc, inc in generation_pairs:
            # Process of obtaining inclusive data for a node differs if the DataFrame has an Index vs a MultiIndex
            if isinstance(self.dataframe.index, pd.MultiIndex):
                new_data = {}
                # Traverse every node in the Graph
                for node in self.graph.traverse():
                    # Consider each unique portion of the MultiIndex corresponding to the current node
                    for non_node_idx in self.dataframe.loc[(node)].index.unique():
                        # If there's only 1 index level besides "node", add it to a 1-element list to ensure consistent typing
                        if not isinstance(non_node_idx, tuple) and not isinstance(
                            non_node_idx, list
                        ):
                            non_node_idx = [non_node_idx]
                        # Build the full index
                        # TODO: Replace the full_idx assignment with the following when 2.7 support
                        # is dropped:
                        # full_idx = (node, *non_node_idx)
                        full_idx = tuple([node]) + tuple(non_node_idx)
                        # Iterate over the children of the current node and add up
                        # their values for the inclusive metric
                        inc_sum = 0
                        for child in node.children:
                            # TODO: See note about full_idx above
                            child_idx = tuple([child]) + tuple(non_node_idx)
                            inc_sum += self.dataframe.loc[child_idx, inc]
                        # Subtract the current node's inclusive metric from the previously calculated sum to
                        # get the exclusive metric value for the node
                        new_data[full_idx] = self.dataframe.loc[full_idx, inc] - inc_sum
                # Add the exclusive metric as a new column in the DataFrame
                self.dataframe = self.dataframe.assign(
                    **{exc: pd.Series(data=new_data)}
                )
            else:
                # Create a basic Node-metric dict for the new exclusive metric
                new_data = {n: -1 for n in self.dataframe.index.values}
                # Traverse the graph
                for node in self.graph.traverse():
                    # Sum up the inclusive metric values of the current node's children
                    inc_sum = 0
                    for child in node.children:
                        inc_sum += self.dataframe.loc[child, inc]
                    # Subtract the current node's inclusive metric from the previously calculated sum to
                    # get the exclusive metric value for the node
                    new_data[node] = self.dataframe.loc[node, inc] - inc_sum
                # Add the exclusive metric as a new column in the DataFrame
                self.dataframe = self.dataframe.assign(
                    **{exc: pd.Series(data=new_data)}
                )
        # Add the newly created metrics to self.exc_metrics
        self.exc_metrics.extend([metric_tuple[0] for metric_tuple in generation_pairs])
        self.exc_metrics = list(set(self.exc_metrics))

    def update_inclusive_columns(self):
        """Update inclusive columns (typically after operations that rewire the
        graph.
        """
        # we should update inc metric only if exc metric exist
        if not self.exc_metrics:
            return

        # TODO When Python 2.7 support is dropped, change this line to the more idiomatic:
        # old_inc_metrics = self.inc_metrics.copy()
        old_inc_metrics = list(self.inc_metrics)
        # TODO Change this logic when inc_metrics and exc_metrics are changed
        new_inc_metrics = []
        for exc in self.exc_metrics:
            if isinstance(exc, tuple):
                if exc[-1].endswith("(exc)"):
                    temp = list(exc)
                    temp[-1] = temp[-1][: -len("(exc)")].strip()
                    new_inc_metrics.append(tuple(temp))
                else:
                    temp = list(exc)
                    temp[-1] = "%s (inc)" % temp[-1]
                    new_inc_metrics.append(tuple(temp))
            else:
                if exc.endswith("(exc)"):
                    new_inc_metrics.append(exc[: -len("(exc)")].strip())
                else:
                    new_inc_metrics.append("%s (inc)" % exc)
        self.inc_metrics = new_inc_metrics

        self.subgraph_sum(self.exc_metrics, self.inc_metrics)
        self.inc_metrics = list(set(self.inc_metrics + old_inc_metrics))

    def show_metric_columns(self):
        """Returns a list of dataframe column labels."""
        return list(self.exc_metrics + self.inc_metrics)

    def unify(self, other):
        """Returns a unified graphframe.

        Ensure self and other have the same graph and same node IDs. This may
        change the node IDs in the dataframe.

        Update the graphs in the graphframe if they differ.
        """
        if self.graph is other.graph:
            return

        node_map = {}
        union_graph = self.graph.union(other.graph, node_map)

        self_index_names = self.dataframe.index.names
        other_index_names = other.dataframe.index.names

        self.dataframe.reset_index(inplace=True)
        other.dataframe.reset_index(inplace=True)

        self.dataframe["node"] = self.dataframe["node"].apply(lambda x: node_map[id(x)])
        other.dataframe["node"] = other.dataframe["node"].apply(
            lambda x: node_map[id(x)]
        )

        # add missing rows to copy of self's dataframe in preparation for
        # operation
        self._insert_missing_rows(other)

        self.dataframe.set_index(self_index_names, inplace=True, drop=True)
        other.dataframe.set_index(other_index_names, inplace=True, drop=True)

        self.graph = union_graph
        other.graph = union_graph

    @deprecated_params(
        metric="metric_column",
        name="name_column",
        expand_names="expand_name",
        context="context_column",
        invert_colors="invert_colormap",
    )
    def tree(
        self,
        metric_column=None,
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
        render_header=True,
    ):
        """Format this graphframe as a tree and return the resulting string."""
        color = sys.stdout.isatty()
        shell = None
        if metric_column is None:
            metric_column = self.default_metric

        if color is False:
            try:
                import IPython

                shell = IPython.get_ipython().__class__.__name__
            except ImportError:
                pass
            # Test if running in a Jupyter notebook or qtconsole
            if shell == "ZMQInteractiveShell":
                color = True

        if sys.version_info.major == 2:
            unicode = False
        elif sys.version_info.major == 3:
            unicode = True

        return ConsoleRenderer(unicode=unicode, color=color).render(
            self.graph.roots,
            self.dataframe,
            metric_column=metric_column,
            precision=precision,
            name_column=name_column,
            expand_name=expand_name,
            context_column=context_column,
            rank=rank,
            thread=thread,
            depth=depth,
            highlight_name=highlight_name,
            colormap=colormap,
            invert_colormap=invert_colormap,
            render_header=render_header,
        )

    def to_dot(self, metric=None, name="name", rank=0, thread=0, threshold=0.0):
        """Write the graph in the graphviz dot format:
        https://www.graphviz.org/doc/info/lang.html
        """
        if metric is None:
            metric = self.default_metric
        return trees_to_dot(
            self.graph.roots, self.dataframe, metric, name, rank, thread, threshold
        )

    def to_flamegraph(self, metric=None, name="name", rank=0, thread=0, threshold=0.0):
        """Write the graph in the folded stack output required by FlameGraph
        http://www.brendangregg.com/flamegraphs.html
        """
        folded_stack = ""
        if metric is None:
            metric = self.default_metric

        for root in self.graph.roots:
            for hnode in root.traverse():
                callpath = hnode.path()
                for i in range(0, len(callpath) - 1):
                    if (
                        "rank" in self.dataframe.index.names
                        and "thread" in self.dataframe.index.names
                    ):
                        df_index = (callpath[i], rank, thread)
                    elif "rank" in self.dataframe.index.names:
                        df_index = (callpath[i], rank)
                    elif "thread" in self.dataframe.index.names:
                        df_index = (callpath[i], thread)
                    else:
                        df_index = callpath[i]
                    folded_stack = (
                        folded_stack + str(self.dataframe.loc[df_index, "name"]) + "; "
                    )

                if (
                    "rank" in self.dataframe.index.names
                    and "thread" in self.dataframe.index.names
                ):
                    df_index = (callpath[-1], rank, thread)
                elif "rank" in self.dataframe.index.names:
                    df_index = (callpath[-1], rank)
                elif "thread" in self.dataframe.index.names:
                    df_index = (callpath[-1], thread)
                else:
                    df_index = callpath[-1]
                folded_stack = (
                    folded_stack + str(self.dataframe.loc[df_index, "name"]) + " "
                )

                # set dataframe index based on if rank and thread are part of the index
                if (
                    "rank" in self.dataframe.index.names
                    and "thread" in self.dataframe.index.names
                ):
                    df_index = (hnode, rank, thread)
                elif "rank" in self.dataframe.index.names:
                    df_index = (hnode, rank)
                elif "thread" in self.dataframe.index.names:
                    df_index = (hnode, thread)
                else:
                    df_index = hnode

                folded_stack = (
                    folded_stack
                    + str(round(self.dataframe.loc[df_index, metric]))
                    + "\n"
                )

        return folded_stack

    def to_literal(self, name="name", rank=0, thread=0, cat_columns=[]):
        """Format this graph as a list of dictionaries for Roundtrip
        visualizations.
        """
        graph_literal = []
        visited = []

        def _get_df_index(hnode):
            if (
                "rank" in self.dataframe.index.names
                and "thread" in self.dataframe.index.names
            ):
                df_index = (hnode, rank, thread)
            elif "rank" in self.dataframe.index.names:
                df_index = (hnode, rank)
            elif "thread" in self.dataframe.index.names:
                df_index = (hnode, thread)
            else:
                df_index = hnode

            return df_index

        def metrics_to_dict(df_index):
            metrics_dict = {}
            for m in sorted(self.inc_metrics + self.exc_metrics):
                node_metric_val = self.dataframe.loc[df_index, m]
                if isinstance(node_metric_val, pd.Series):
                    node_metric_val = node_metric_val[0]
                if np.isinf(node_metric_val) or np.isneginf(node_metric_val):
                    node_metric_val = 0.0
                if pd.isna(node_metric_val):
                    node_metric_val = 0.0
                metrics_dict[m] = node_metric_val

            return metrics_dict

        def attributes_to_dict(df_index):
            valid_columns = [
                col for col in cat_columns if col in self.dataframe.columns
            ]

            attributes_dict = {}
            for m in sorted(valid_columns):
                node_attr_val = self.dataframe.loc[df_index, m]
                if isinstance(node_attr_val, pd.Series):
                    node_attr_val = node_attr_val[0]
                attributes_dict[m] = node_attr_val

            return attributes_dict

        def add_nodes(hnode):
            df_index = _get_df_index(hnode)

            node_dict = {}

            node_name = self.dataframe.loc[df_index, name]

            if isinstance(node_name, pd.Series):
                self.dataframe.loc[df_index]
                node_name = node_name[0]

            node_dict["name"] = node_name
            node_dict["frame"] = hnode.frame.attrs
            node_dict["metrics"] = metrics_to_dict(df_index)
            # node_dict["metrics"]["_hatchet_nid"] = int(self.dataframe["nid"][df_index])
            node_dict["metrics"]["_hatchet_nid"] = int(hnode._hatchet_nid)
            node_dict["attributes"] = attributes_to_dict(df_index)

            if hnode.children and hnode not in visited:
                visited.append(hnode)
                node_dict["children"] = []

                for child in sorted(hnode.children, key=lambda n: n.frame):
                    node_dict["children"].append(add_nodes(child))

            return node_dict

        for root in sorted(self.graph.roots, key=lambda n: n.frame):
            graph_literal.append(add_nodes(root))

        return graph_literal

    def to_dict(self):
        hatchet_dict = {}

        """
        Nodes: {hatchet_nid: {node data, children:[by-id]}}
        """
        graphs = []
        for root in self.graph.roots:
            formatted_graph_dict = {}
            for n in root.traverse():
                formatted_graph_dict[n._hatchet_nid] = {
                    "data": n.frame.attrs,
                    "children": [c._hatchet_nid for c in n.children],
                }
            graphs.append(formatted_graph_dict)

        hatchet_dict["graph"] = graphs

        hatchet_dict["dataframe_indices"] = list(self.dataframe.index.names)
        ef = self.dataframe.reset_index()
        ef["node"] = ef["node"].apply(lambda n: n._hatchet_nid)
        hatchet_dict["dataframe"] = ef.replace({np.nan: None}).to_dict("records")

        hatchet_dict["inclusive_metrics"] = self.inc_metrics
        hatchet_dict["exclusive_metrics"] = self.exc_metrics

        return hatchet_dict

    def to_json(self):
        return json.dumps(self.to_dict())

    def _operator(self, other, op):
        """Generic function to apply operator to two dataframes and store
        result in self.

        Arguments:
            self (graphframe): self's graphframe
            other (graphframe): other's graphframe
            op (operator): pandas arithmetic operator

        Return:
            (GraphFrame): self's graphframe modified
        """
        # unioned set of self and other exclusive and inclusive metrics
        all_metrics = list(
            set().union(
                self.exc_metrics, self.inc_metrics, other.exc_metrics, other.inc_metrics
            )
        )

        self.dataframe.update(op(other.dataframe[all_metrics]))

        return self

    def _insert_missing_rows(self, other):
        """Helper function to add rows that exist in other, but not in self.

        This returns a graphframe with a modified dataframe. The new rows will
        contain zeros for numeric columns.

        Return:
            (GraphFrame): self's modified graphframe
        """
        all_metrics = list(
            set().union(
                self.exc_metrics, self.inc_metrics, other.exc_metrics, other.inc_metrics
            )
        )

        # make two 2D nparrays arrays with two columns:
        # 1) the hashed value of a node and 2) a numerical index
        # Many operations are stacked here to reduce the need for storing
        # large intermediary datasets
        self_hsh_ndx = np.vstack(
            (
                np.array(
                    [x.__hash__() for x in self.dataframe["node"]], dtype=np.uint64
                ),
                self.dataframe.index.values.astype(np.uint64),
            )
        ).T
        other_hsh_ndx = np.vstack(
            (
                np.array(
                    [x.__hash__() for x in other.dataframe["node"]], dtype=np.uint64
                ),
                other.dataframe.index.values.astype(np.uint64),
            )
        ).T

        # sort our 2D arrays by hashed node value so a binary search can be used
        # in the cython function fast_not_isin
        self_hsh_ndx_sorted = self_hsh_ndx[self_hsh_ndx[:, 0].argsort()]
        other_hsh_ndx_sorted = other_hsh_ndx[other_hsh_ndx[:, 0].argsort()]

        # get nodes that exist in other, but not in self, set metric columns to 0 for
        # these rows
        other_not_in_self = other.dataframe[
            _gfm_cy.fast_not_isin(
                other_hsh_ndx_sorted,
                self_hsh_ndx_sorted,
                other_hsh_ndx_sorted.shape[0],
                self_hsh_ndx_sorted.shape[0],
            )
        ]
        # get nodes that exist in self, but not in other
        self_not_in_other = self.dataframe[
            _gfm_cy.fast_not_isin(
                self_hsh_ndx_sorted,
                other_hsh_ndx_sorted,
                self_hsh_ndx_sorted.shape[0],
                other_hsh_ndx_sorted.shape[0],
            )
        ]

        # if there are missing nodes in either self or other, add a new column
        # called _missing_node
        if not self_not_in_other.empty:
            self.dataframe = self.dataframe.assign(
                _missing_node=np.zeros(len(self.dataframe), dtype=np.short)
            )
        if not other_not_in_self.empty:
            # initialize with 2 to save filling in later
            other_not_in_self = other_not_in_self.assign(
                _missing_node=[int(2) for x in range(len(other_not_in_self))]
            )

            # add a new column to self if other has nodes not in self
            if self_not_in_other.empty:
                self.dataframe["_missing_node"] = np.zeros(
                    len(self.dataframe), dtype=np.short
                )

        # get lengths to pass into
        onis_len = len(other_not_in_self)
        snio_len = len(self_not_in_other)

        # case where self is a superset of other
        if snio_len != 0:
            self_missing_node = self.dataframe["_missing_node"].values
            snio_indices = self_not_in_other.index.values

            # This function adds 1 to all nodes in self.dataframe['_missing_node'] which
            # are in self but not in the other graphframe
            _gfm_cy.insert_one_for_self_nodes(snio_len, self_missing_node, snio_indices)
            self.dataframe["_missing_node"] = np.array(
                [n for n in self_missing_node], dtype=np.short
            )

        # for nodes that only exist in other, set the metric to be nan (since
        # it's a missing node in self)
        # replaces individual metric assignments with np.zeros
        for j in all_metrics:
            other_not_in_self[j] = np.full(onis_len, np.nan)

        # append missing rows (nodes that exist in other, but not in self) to self's
        # dataframe
        self.dataframe = pd.concat(
            [self.dataframe, other_not_in_self], axis=0, sort=True
        )

        return self

    def groupby_aggregate(self, groupby_function, agg_function):
        """Groupby-aggregate dataframe and reindex the Graph.

        Reindex the graph to match the groupby-aggregated dataframe.

        Update the frame attributes to contain those columns in the dataframe index.

        Arguments:
            self (graphframe): self's graphframe
            groupby_function: groupby function on dataframe
            agg_function: aggregate function on dataframe

        Return:
            (GraphFrame): new graphframe with reindexed graph and groupby-aggregated dataframe
        """
        # create new nodes for each unique node in the old dataframe
        # length is equal to number of nodes in original graph
        old_to_new = {}

        # list of new roots
        new_roots = []

        # dict of (new) super nodes
        # length is equal to length of dataframe index (after groupby-aggregate)
        node_dicts = []

        def reindex(node, parent, visited):
            """Reindex the graph.

            Connect super nodes to children according to relationships from old graph.
            """
            # grab the super node corresponding to original node
            super_node = old_to_new.get(node)

            if not node.parents and super_node not in new_roots:
                # this is a new root
                new_roots.append(super_node)

            # iterate over parents of old node, adding parents to super node
            for parent in node.parents:
                # convert node to super node
                snode = old_to_new.get(parent)
                # move to next node if parent and super node are to be merged
                if snode == super_node:
                    continue
                # add node to super node's parents if parent does not exist in super
                # node's parents
                if snode not in super_node.parents:
                    super_node.add_parent(snode)

            # iterate over children of old node, adding children to super node
            for child in node.children:
                # convert node to super node
                snode = old_to_new.get(child)
                # move to next node if child and super node are to be merged
                if snode == super_node:
                    continue
                # add node to super node's children if child does not exist in super
                # node's children
                if snode not in super_node.children:
                    super_node.add_child(snode)

            if node not in visited:
                visited.add(node)
                for child in node.children:
                    reindex(child, super_node, visited)

        # groupby-aggregate dataframe based on user-supplied functions
        groupby_obj = self.dataframe.groupby(groupby_function)
        agg_df = groupby_obj.agg(agg_function)

        # traverse groupby_obj, determine old node to super node mapping
        nid = 0
        for k, v in groupby_obj.groups.items():
            node_name = k
            node_type = agg_df.index.name
            super_node = Node(Frame({"name": node_name, "type": node_type}), None, nid)
            n = {"node": super_node, "nid": nid, "name": node_name}
            node_dicts.append(n)
            nid += 1

            # if many old nodes map to the same super node
            for i in v:
                old_to_new[i] = super_node

        # reindex graph by traversing old graph
        visited = set()
        for root in self.graph.roots:
            reindex(root, None, visited)

        # append super nodes to groupby-aggregate dataframe
        df_index = list(agg_df.index.names)
        agg_df.reset_index(inplace=True)
        df_nodes = pd.DataFrame.from_dict(data=node_dicts)
        tmp_df = pd.concat([agg_df, df_nodes], axis=1)
        # add node to dataframe index if it doesn't exist
        if "node" not in df_index:
            df_index.append("node")
        # reset index
        tmp_df.set_index(df_index, inplace=True)

        # update _hatchet_nid in reindexed graph and groupby-aggregate dataframe
        graph = Graph(new_roots)
        graph.enumerate_traverse()

        # put it all together
        new_gf = GraphFrame(
            graph,
            tmp_df,
            self.exc_metrics,
            self.inc_metrics,
            self.default_metric,
            self.metadata,
        )
        new_gf.drop_index_levels()
        return new_gf

    def add(self, other):
        """Returns the column-wise sum of two graphframes as a new graphframe.

        This graphframe is the union of self's and other's graphs, and does not
        modify self or other.

        Return:
            (GraphFrame): new graphframe
        """
        # create a copy of both graphframes
        self_copy = self.copy()
        other_copy = other.copy()

        # unify copies of graphframes
        self_copy.unify(other_copy)

        return self_copy._operator(other_copy, self_copy.dataframe.add)

    def sub(self, other):
        """Returns the column-wise difference of two graphframes as a new
        graphframe.

        This graphframe is the union of self's and other's graphs, and does not
        modify self or other.

        Return:
            (GraphFrame): new graphframe
        """
        # create a copy of both graphframes
        self_copy = self.copy()
        other_copy = other.copy()

        # unify copies of graphframes
        self_copy.unify(other_copy)

        return self_copy._operator(other_copy, self_copy.dataframe.sub)

    def div(self, other):
        """Returns the column-wise float division of two graphframes as a new graphframe.

        This graphframe is the union of self's and other's graphs, and does not
        modify self or other.

        Return:
            (GraphFrame): new graphframe
        """
        # create a copy of both graphframes
        self_copy = self.copy()
        other_copy = other.copy()

        # unify copies of graphframes
        self_copy.unify(other_copy)

        return self_copy._operator(other_copy, self_copy.dataframe.divide)

    def mul(self, other):
        """Returns the column-wise float multiplication of two graphframes as a new graphframe.

        This graphframe is the union of self's and other's graphs, and does not
        modify self or other.

        Return:
            (GraphFrame): new graphframe
        """
        # create a copy of both graphframes
        self_copy = self.copy()
        other_copy = other.copy()

        # unify copies of graphframes
        self_copy.unify(other_copy)

        return self_copy._operator(other_copy, self_copy.dataframe.multiply)

    def __iadd__(self, other):
        """Computes column-wise sum of two graphframes and stores the result in
        self.

        Self's graphframe is the union of self's and other's graphs, and the
        node handles from self will be rewritten with this operation. This
        operation does not modify other.

        Return:
            (GraphFrame): self's graphframe modified
        """
        # create a copy of other's graphframe
        other_copy = other.copy()

        # unify self graphframe and copy of other graphframe
        self.unify(other_copy)

        return self._operator(other_copy, self.dataframe.add)

    def __add__(self, other):
        """Returns the column-wise sum of two graphframes as a new graphframe.

        This graphframe is the union of self's and other's graphs, and does not
        modify self or other.

        Return:
            (GraphFrame): new graphframe
        """
        return self.add(other)

    def __mul__(self, other):
        """Returns the column-wise multiplication of two graphframes as a new graphframe.

        This graphframe is the union of self's and other's graphs, and does not
        modify self or other.

        Return:
            (GraphFrame): new graphframe
        """
        return self.mul(other)

    def __isub__(self, other):
        """Computes column-wise difference of two graphframes and stores the
        result in self.

        Self's graphframe is the union of self's and other's graphs, and the
        node handles from self will be rewritten with this operation. This
        operation does not modify other.

        Return:
            (GraphFrame): self's graphframe modified
        """
        # create a copy of other's graphframe
        other_copy = other.copy()

        # unify self graphframe and other graphframe
        self.unify(other_copy)

        return self._operator(other_copy, self.dataframe.sub)

    def __sub__(self, other):
        """Returns the column-wise difference of two graphframes as a new
        graphframe.

        This graphframe is the union of self's and other's graphs, and does not
        modify self or other.

        Return:
            (GraphFrame): new graphframe
        """
        return self.sub(other)

    def __idiv__(self, other):
        """Computes column-wise float division of two graphframes and stores the
        result in self.

        Self's graphframe is the union of self's and other's graphs, and the
        node handles from self will be rewritten with this operation. This
        operation does not modify other.

        Return:
            (GraphFrame): self's graphframe modified
        """
        # create a copy of other's graphframe
        other_copy = other.copy()

        # unify self graphframe and other graphframe
        self.unify(other_copy)

        return self._operator(other_copy, self.dataframe.div)

    def __truediv__(self, other):
        """Returns the column-wise float division of two graphframes as a new
        graphframe.

        This graphframe is the union of self's and other's graphs, and does not
        modify self or other.

        Return:
            (GraphFrame): new graphframe
        """
        return self.div(other)

    def __imul__(self, other):
        """Computes column-wise float multiplication of two graphframes and stores the
        result in self.

        Self's graphframe is the union of self's and other's graphs, and the
        node handles from self will be rewritten with this operation. This
        operation does not modify other.

        Return:
            (GraphFrame): self's graphframe modified
        """
        # create a copy of other's graphframe
        other_copy = other.copy()

        # unify self graphframe and other graphframe
        self.unify(other_copy)

        return self._operator(other_copy, self.dataframe.mul)


class InvalidFilter(Exception):
    """Raised when an invalid argument is passed to the filter function."""


class EmptyFilter(Exception):
    """Raised when a filter would otherwise return an empty GraphFrame."""



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from bs4 import BeautifulSoup
from os import walk, path

static_filepath = path.abspath("static/")

for (pt, dirs, files) in walk(static_filepath):
    for file in files:
        if ".html" in file:
            with open(path.join(static_filepath, file), "r") as f:
                html = f.read()
                soup = BeautifulSoup(html)
                soup.script["src"] = path.join(static_filepath, file[0:-5] + ".js")
                with open(path.join(static_filepath, file), "w") as writer:
                    writer.write(soup.prettify())



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import warnings
import sys

from .dataframe_writer import DataframeWriter


class HDF5Writer(DataframeWriter):
    def __init__(self, filename):
        if sys.version_info[0] == 2:
            super(HDF5Writer, self).__init__(filename)
        else:
            super().__init__(filename)

    def _write_dataframe_to_file(self, df, **kwargs):
        if "key" not in kwargs:
            raise KeyError("Writing to HDF5 requires a user-supplied key")
        key = kwargs["key"]
        del kwargs["key"]
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=Warning)
            df.to_hdf(self.filename, key, **kwargs)



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import json
import pandas as pd
import os
import glob
from hatchet.graphframe import GraphFrame
from ..node import Node
from ..graph import Graph
from ..frame import Frame
from ..util.timer import Timer


class TimemoryReader:
    """Read in timemory JSON output"""

    def __init__(self, input, select=None, **_kwargs):
        """Arguments:
        input (str or file-stream or dict or None):
            Valid argument types are:

            1. Filename for a timemory JSON tree file
            2. Open file stream to one of these files
            3. Dictionary from timemory JSON tree

        select (list of str):
            A list of strings which match the component enumeration names, e.g. ["cpu_clock"].

        per_thread (boolean):
            Ensures that when applying filters to the graphframe, frames with
            identical name/file/line/etc. info but from different threads are not
            combined

        per_rank (boolean):
            Ensures that when applying filters to the graphframe, frames with
            identical name/file/line/etc. info but from different ranks are not
            combined
        """
        self.graph_dict = {"timemory": {}}
        self.input = input
        self.default_metric = None
        self.timer = Timer()
        self.metric_cols = []
        self.properties = {}
        self.include_tid = True
        self.include_nid = True
        self.multiple_ranks = False
        self.multiple_threads = False
        self.callpath_to_node_dict = {}  # (callpath, rank, thread): <node_dict>
        self.callpath_to_node = {}  # (callpath): <node>

        # the per_thread and per_rank settings make sure that
        # squashing doesn't collapse the threads/ranks
        self.per_thread = _kwargs["per_thread"] if "per_thread" in _kwargs else False
        self.per_rank = _kwargs["per_rank"] if "per_rank" in _kwargs else False

        if select is None:
            self.select = select
        elif isinstance(select, list):
            if select:

                def _get_select(val):
                    if isinstance(val, str):
                        return val.lower()
                    elif callable(val):
                        return val().lower()
                    raise TypeError(
                        "Items in select must be string or callable: {}".format(
                            type(val).__name__
                        )
                    )

                self.select = [_get_select(v) for v in select]
        else:
            raise TypeError("select must be None or list of string")

    def create_graph(self):
        """Create graph and dataframe"""
        list_roots = []

        def remove_keys(_dict, _keys):
            """Remove keys from dictionary"""
            if isinstance(_keys, str):
                if _keys in _dict:
                    del _dict[_keys]
            else:
                for _key in _keys:
                    _dict = remove_keys(_dict, _key)
            return _dict

        def add_metrics(_dict):
            """Add any keys to metric_cols which don't already exist"""
            for key, itr in _dict.items():
                if key not in self.metric_cols:
                    self.metric_cols.append(key)

        def process_regex(_data):
            """Process the regex data for func/file/line info"""
            _tmp = {}
            if _data is not None and len(_data.groups()) > 0:
                for _key in ("head", "func", "file", "line", "tail"):
                    try:
                        _val = _data.group(_key)
                        if _val:
                            _tmp[_key] = _val
                    except Exception:
                        pass
            return _tmp if _tmp else None

        def perform_regex(_prefix):
            """Performs a search for standard configurations of function + file + line"""
            import re

            _tmp = None
            for _pattern in [
                # [func][file]
                r"(^\[)(?P<func>.*)(\]\[)(?P<file>.*)(\]$)",
                # label  [func/file:line]
                r"(?P<head>.+)([ \t]+)\[(?P<func>\S+)([/])(?P<file>\S+):(?P<line>[0-9]+)\]$",
                # func@file:line/tail
                # func/file:line/tail
                r"(?P<func>\S+)([@/])(?P<file>\S+):(?P<line>[0-9]+)[/]*(?P<tail>.*)",
                # func@file/tail
                # func/file/tail
                r"(?P<func>\S+)([@/])(?P<file>\S+)([/])(?P<tail>.*)",
                # func:line/tail
                r"(?P<func>\S+):(?P<line>[0-9]+)([/]*)(?P<tail>.*)",
            ]:
                _tmp = process_regex(re.search(_pattern, _prefix))
                if _tmp:
                    break
            return _tmp if _tmp else None

        def get_name_line_file(_prefix):
            """Get the standard set of dictionary entries.
            Also, parses the prefix for func-file-line info
            which is typically in the form:
                <FUNC>@<FILE>:<LINE>/...
                <FUNC>/<FILE>:<LINE>/...
                <SOURCE>    [<FUNC>/<FILE>:<LINE>]
            """
            _keys = {
                "type": "region",
                "name": _prefix,
            }
            _extra = {"file": "<unknown>", "line": "0"}
            _pdict = perform_regex(_prefix)
            if _pdict is not None:
                if "head" in _pdict:
                    _keys["name"] = _pdict["head"].rstrip()
                    _extra["line"] = _pdict["line"]
                    _extra["file"] = _pdict["file"]
                else:
                    _keys["name"] = _pdict["func"]
                    _extra["file"] = (
                        _pdict["file"] if "file" in _pdict else "<unknown file>"
                    )
                    if "line" in _pdict:
                        _extra["line"] = _pdict["line"]
                    if "tail" in _pdict:
                        _keys["name"] = "{}/{}".format(_keys["name"], _pdict["tail"])
            return (_keys, _extra)

        def format_labels(_labels):
            """Formats multi dimensional metrics which refer to multiple metrics
            stored in a 1D list.

            Example: PAPI_TOT_CYC, PAPI_TOT_INS, and PAPI_L2_TCM are stored as
            ["Total_cycles", "Instr_completed", "L2_cache_misses"].

            After formatting:
            ['Total-cycles', 'Instr-completed', 'L2-cache-misses']
            """
            _ret = []
            if isinstance(_labels, str):
                # put in a list if the label is a string.
                _ret = [_labels]
            elif isinstance(_labels, dict):
                for _key, _item in _labels.items():
                    _ret.append(_key.strip().replace(" ", "-").replace("_", "-"))
            elif isinstance(_labels, list) or isinstance(_labels, tuple):
                for _item in _labels:
                    _ret.append(_item.strip().replace(" ", "-").replace("_", "-"))
            return _ret

        def match_labels_and_values(_metric_stats, _metric_label, _metric_type):
            """Match metric labels with values and add '(inc)' if the metric type
            is inclusive.

            _metric_stat example 1: {'sum': 0.010, 'min': 0.001, ...}
            _metric_stat example 2: {'sum': [0.010, 0.020, 0.030], ...}

            _metric_label example 1: wall_clock
            _metric_label example 2: ['Total-cycles', 'Instr-completed', 'L2-cache-misses']

            _metric_type: ' (inc)' or ''
            """
            _ret = {}
            for _key, _item in _metric_stats.items():
                if isinstance(_item, dict):
                    for i, (k, v) in enumerate(_item.items()):
                        _ret["{}.{}{}".format(_key, _metric_label[i], _metric_type)] = v
                # match with metric labels if _metric_stat item is a list.
                elif isinstance(_item, list):
                    for i in range(len(_item)):
                        _ret[
                            "{}.{}{}".format(_key, _metric_label[i], _metric_type)
                        ] = _item[i]
                # check if _metric_stat item is not a dict or list
                else:
                    _ret["{}.{}{}".format(_key, _metric_label, _metric_type)] = _item
            return _ret

        def collapse_ids(_obj, _expect_scalar=False):
            """node/rank/thread id may be int, array of ints, or None.
            When the entry is a list of integers (which happens when metric values
            are aggregates of multiple ranks/threads), this function generates a consistent
            form which is NOT numerical to avoid `groupby(...).sum()` operations from producing
            something nonsensical (i.e. adding together thread-ids) but ensures the entry is
            still hashable (i.e. a list of integers is not hashable and will cause `groupby(...).sum()` to
            throw an error)

            Arguments:
                _obj (int or list of ints):
                    The node/rank/thread id(s) for the metric.
                    If a list is provided, this implies that the metric values are aggregations from multiple nodes/ranks/threads

                _expect_scalar (bool):
                    Setting this value to true means that `_obj` is expected to be an integer and the
                    return value should be converted to an integer. If this value is true and an array of ints
                    is passed, an error will be thrown

            Return Value:
                if _expect_scalar is False: string
                if _expect_scalar is True: int
            """
            if isinstance(_obj, list):
                if len(_obj) == 1:
                    return int(_obj[0])
                else:
                    if _expect_scalar:
                        raise ValueError(
                            f"collapse_ids expected per-rank or per-thread values but list of ids ({_obj}) implies that data is aggregated across multiple ranks or threads"
                        )
                    return ",".join([f"{x}" for x in _obj]).strip(",")
            elif _obj is not None:
                return f"{_obj}" if _expect_scalar else int(_obj)
            return None

        def parse_node(_metric_name, _node_data, _hparent, _rank, _parent_callpath):
            """Create callpath_to_node_dict for one node and then call the function
            recursively on all children.
            """

            # If the hash is zero, that indicates that the node
            # is a dummy for the root or is used for synchronizing data
            # between multiple threads
            # TODO: do we have some intermediate nodes that have hash = 0?
            if _node_data["node"]["hash"] == 0:
                if "children" in _node_data:
                    for _child in _node_data["children"]:
                        parse_node(
                            _metric_name,
                            _child,
                            _hparent,
                            _rank,
                            _parent_callpath,
                        )
                return

            _prop = self.properties[_metric_name]
            _frame_attrs, _extra = get_name_line_file(_node_data["node"]["prefix"])

            callpath = _parent_callpath + (_frame_attrs["name"],)

            # check if the node already exits.
            _hnode = self.callpath_to_node.get(callpath)
            if _hnode is None:
                # connect with the parent during node creation.
                _hnode = Node(Frame(_frame_attrs), _hparent)
                self.callpath_to_node[callpath] = _hnode
                if _hparent is None:
                    # if parent is none, this is a root node.
                    list_roots.append(_hnode)
                else:
                    # if parent is not none, add as a child.
                    _hparent.add_child(_hnode)

            # by placing the thread-id or rank-id in _frame_attrs, the hash
            # for the Frame(_keys) effectively circumvent Hatchet's
            # default behavior of combining similar thread/rank entries
            _tid_dict = _frame_attrs if self.per_thread else _extra
            _rank_dict = _frame_attrs if self.per_rank else _extra

            # handle the rank
            _rank_dict["rank"] = collapse_ids(_rank, self.per_rank)
            if _rank_dict["rank"] is None:
                del _rank_dict["rank"]
                self.include_nid = False

            # extract some relevant data
            _tid_dict["thread"] = collapse_ids(
                _node_data["node"]["tid"], self.per_thread
            )
            _extra["pid"] = collapse_ids(_node_data["node"]["pid"], False)
            _extra["count"] = _node_data["node"]["inclusive"]["entry"]["laps"]

            # check if there are multiple threads
            # TODO: move this outside if don't have per thread data in timemory
            if not self.multiple_threads:
                if _tid_dict["thread"] != 0:
                    self.multiple_threads = True

            # this is the name for the metrics
            _labels = None if "type" not in _prop else _prop["type"]
            # if the labels are not a single string, they are multi-dimensional
            _metrics_in_vector = True if not isinstance(_labels, str) else False
            # remove some spurious data from inclusive/exclusive stats
            _remove = ["cereal_class_version", "count"]
            _inc_stats = remove_keys(_node_data["node"]["inclusive"]["stats"], _remove)
            _exc_stats = remove_keys(_node_data["node"]["exclusive"]["stats"], _remove)

            # if multi-dimensions, create alternative "sum.<...>", etc. labels + data
            # add " (inc)" to the end of every column that represents an inclusive stat
            if _metrics_in_vector:
                # Example of a multi-dimensional output: if we have 3 papi events
                # PAPI_TOT_CYC, PAPI_TOT_INS, PAPI_L2_TCM:
                # _metric_labels: ["Total_cycles", "Instr_completed", "L2_cache_misses"]
                # _exc_stats -> "sum": [8301.0, 4910.0, 275.0],
                _metric_labels = format_labels(_labels)
                _exc_stats = match_labels_and_values(_exc_stats, _metric_labels, "")
                _inc_stats = match_labels_and_values(
                    _inc_stats, _metric_labels, " (inc)"
                )
            else:
                # add metric name and type.
                # Example: sum -> sum.wall_clock (inc)
                _inc_stats = match_labels_and_values(_inc_stats, _metric_name, " (inc)")
                # Example: sum -> sum.wallclock
                _exc_stats = match_labels_and_values(_exc_stats, _metric_name, "")

            # add the inclusive and exclusive columns to the list of relevant column names
            add_metrics(_exc_stats)
            add_metrics(_inc_stats)

            # we use callpath_to_node_dict instead of directly
            # using node_dicts to be able to merge metrics.
            # We use its values later as node_dicts.
            # (callpath, rank, thread): <node_dict>
            callpath_rank_thread = tuple((callpath, _rank, _tid_dict["thread"]))
            node_dict = self.callpath_to_node_dict.get(callpath_rank_thread)
            # check if we saw this (callpath, rank, thread) before
            if node_dict is None:
                # if no, create a new dict.
                self.callpath_to_node_dict[callpath_rank_thread] = dict(
                    {"node": _hnode, **_frame_attrs},
                    **_extra,
                    **_exc_stats,
                    **_inc_stats,
                )
            else:
                # if yes, don't create a new dict, just add the new metrics to
                # the existing node_dict using update().
                # we are doing this to combine different metrics on a single dataframe.
                self.callpath_to_node_dict[callpath_rank_thread].update(
                    dict(**_exc_stats, **_inc_stats)
                )

            # recursion
            if "children" in _node_data:
                for _child in _node_data["children"]:
                    parse_node(_metric_name, _child, _hnode, _rank, callpath)

        def read_graph(_metric_name, ranks_data, _rank):
            """The layout of the graph at this stage
            is subject to slightly different structures
            based on whether distributed memory parallelism (DMP)
            (e.g. MPI, UPC++) was supported and active

            Returns the last rank (_idx).
            """

            rank = None
            total_ranks = len(ranks_data)

            for i in range(total_ranks):
                # rank_data stores all the graph/cct data of a rank
                # starting from the first node in the cct.
                rank_data = ranks_data[i]
                rank = None if _rank is None else i + _rank
                if isinstance(rank_data, list):
                    for data in rank_data:
                        if len(data["children"]) != 0:
                            # empty tuple represents the parent callpath for the root node.
                            # third parameter is the parent node. It's none for the root node.
                            parse_node(
                                _metric_name,
                                data,
                                None,
                                rank,
                                tuple(),
                            )
                else:
                    if len(rank_data["children"]) != 0:
                        # empty tuple represents the parent callpath for the root node.
                        # third parameter is the parent node. It's none for the root node.
                        parse_node(_metric_name, rank_data, None, rank, tuple())

            if total_ranks > 0:
                return True
            return False

        def read_properties(properties, _metric_name, _metric_data):
            """Read in the properties for a component. This
            contains information on the type of the component,
            a description, a unit_value relative to the
            standard, a unit label, whether the data is
            only relevant per-thread, the number of MPI and/or
            UPC++ ranks (some results can theoretically use
            both UPC++ and MPI), the number of threads in
            the application, and the total number of processes
            """
            if _metric_name not in properties:
                properties[_metric_name] = {}
            try:
                properties[_metric_name]["properties"] = remove_keys(
                    _metric_data["properties"], "cereal_class_version"
                )
            except KeyError:
                pass
            for p in (
                "type",
                "description",
                "unit_value",
                "unit_repr",
                "thread_scope_only",
                "mpi_size",
                "upcxx_size",
                "thread_count",
                "process_count",
            ):
                if (
                    p not in properties[_metric_name]
                    or properties[_metric_name][p] is None
                ):
                    if p in _metric_data:
                        properties[_metric_name][p] = _metric_data[p]
                    else:
                        properties[_metric_name][p] = None

        # graph_dict[timemory] stores all metric data.
        # each metric data is another item in this dict.
        for metric_name, metric_data in self.graph_dict["timemory"].items():
            # strip out the namespace if provided
            metric_name = (
                metric_name.replace("tim::", "").replace("component::", "").lower()
            )
            # check for selection
            if self.select is not None and metric_name not in self.select:
                continue
            # read in properties
            read_properties(self.properties, metric_name, metric_data)
            # if no DMP supported
            if "graph" in metric_data:
                read_graph(metric_name, metric_data["graph"], None)
            else:
                # read in MPI results
                if "mpi" in metric_data:
                    self.multiple_ranks = read_graph(metric_name, metric_data["mpi"], 0)
                # if MPI and UPC++, report ranks
                # offset by MPI_Comm_size
                _rank = self.properties[metric_name]["mpi_size"]
                _rank = 0 if _rank is None else int(_rank)
                if "upc" in metric_data:
                    self.multiple_ranks = read_graph(
                        metric_name, metric_data["upc"], _rank
                    )
                elif "upcxx" in metric_data:
                    self.multiple_ranks = read_graph(
                        metric_name, metric_data["upcxx"], _rank
                    )

        # create the graph of the roots
        graph = Graph(list_roots)
        graph.enumerate_traverse()

        # separate out the inclusive vs. exclusive columns
        exc_metrics = []
        inc_metrics = []
        for column in self.metric_cols:
            if column.endswith(" (inc)"):
                inc_metrics.append(column)
            else:
                exc_metrics.append(column)

        # set the default metric
        if self.default_metric is None:
            if len(exc_metrics) > 0:
                if "sum.wall_clock" in exc_metrics:
                    self.default_metric = "sum.wall_clock"
                elif "sum.cpu_clock" in exc_metrics:
                    self.default_metric = "sum.cpu_clock"
                else:
                    self.default_metric = exc_metrics[0]
            elif len(inc_metrics) > 0:
                self.default_metric = inc_metrics[0]
            else:
                self.default_metric = "sum"

        node_dicts = list(self.callpath_to_node_dict.values())
        dataframe = pd.DataFrame(data=node_dicts)

        indices = []
        # Set indices according to rank/thread numbers.
        if self.multiple_ranks and self.multiple_threads:
            indices = ["node", "rank", "thread"]
        elif self.multiple_ranks:
            dataframe.drop(columns=["thread"], inplace=True)
            indices = ["node", "rank"]
        elif self.multiple_threads:
            dataframe.drop(columns=["rank"], inplace=True)
            indices = ["node", "thread"]
        else:
            indices = ["node"]

        dataframe.set_index(indices, inplace=True)
        dataframe.sort_index(inplace=True)

        # Fill the missing ranks
        # After unstacking and iterating over rows, there
        # will be "NaN" values for some ranks. Find the first
        # rank that has notna value and use it for other rows/ranks
        # of the multiindex.
        # TODO: iterrows() is not the best way to iterate over rows.
        if self.multiple_ranks or self.multiple_threads:
            dataframe = dataframe.unstack()
            for idx, row in dataframe.iterrows():

                # There is always a valid name for an index.
                # Take that valid name and assign to other ranks/rows.
                name = row["name"][row["name"].first_valid_index()]
                dataframe.loc[idx, "name"] = name

                # Sometimes there is no file information.
                if row["file"].first_valid_index() is not None:
                    file = row["file"][row["file"].first_valid_index()]
                    dataframe.loc[idx, "file"] = file

                # Sometimes there is no file information.
                if row["type"].first_valid_index() is not None:
                    file = row["type"][row["type"].first_valid_index()]
                    dataframe.loc[idx, "type"] = file

                # Fill the rest with 0
                dataframe.fillna(0, inplace=True)

            # Stack the dataframe
            dataframe = dataframe.stack()

        return GraphFrame(
            graph, dataframe, exc_metrics, inc_metrics, self.default_metric
        )

    def read(self):
        """Read timemory json."""

        # check if the input is a dictionary.
        if isinstance(self.input, dict):
            self.graph_dict = self.input
        # check if the input is a directory and get '.tree.json' files if true.
        elif os.path.isdir(self.input):
            tree_files = glob.glob(self.input + "/*.tree.json")
            for file in tree_files:
                # read all files that end with .tree.json.
                with open(file, "r") as f:
                    # add all metrics to the same dict even though timemory
                    # creates a separate file for each metric.
                    self.graph_dict["timemory"].update(json.load(f)["timemory"])
        # check if the input is a filename that ends in json
        elif isinstance(self.input, str) and self.input.endswith("json"):
            with open(self.input, "r") as f:
                self.graph_dict = json.load(f)
        elif not isinstance(self.input, str):
            self.graph_dict = json.loads(self.input.read())
        else:
            raise TypeError("input must be dict, directory, json file, or string")

        return self.create_graph()



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from numbers import Real
import numpy as np
import pandas as pd
from pandas.api.types import (
    is_numeric_dtype,
    is_string_dtype,
)
import re
import sys

from .errors import InvalidQueryPath, InvalidQueryFilter, MultiIndexModeMismatch
from .query import Query


def _process_multi_index_mode(apply_result, multi_index_mode):
    if multi_index_mode == "any":
        return apply_result.any()
    if multi_index_mode == "all":
        return apply_result.all()
    raise ValueError(
        "Multi-Index Mode for the Object-based dialect must be either 'any' or 'all'"
    )


def _process_predicate(attr_filter, multi_index_mode):
    """Converts high-level API attribute filter to a lambda"""
    compops = ("<", ">", "==", ">=", "<=", "<>", "!=")  # ,

    def filter_series(df_row):
        def filter_single_series(df_row, key, single_value):
            if key == "depth":
                node = df_row.name
                if isinstance(single_value, str) and single_value.lower().startswith(
                    compops
                ):
                    return eval("{} {}".format(node._depth, single_value))
                if isinstance(single_value, Real):
                    # If the value for "depth" is -1, check if the node is a leaf
                    if single_value == -1:
                        return len(node.children) == 0
                    return node._depth == single_value
                raise InvalidQueryFilter(
                    "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                        key
                    )
                )
            if key == "node_id":
                node = df_row.name
                if isinstance(single_value, str) and single_value.lower().startswith(
                    compops
                ):
                    return eval("{} {}".format(node._hatchet_nid, single_value))
                if isinstance(single_value, Real):
                    return node._hatchet_nid == single_value
                raise InvalidQueryFilter(
                    "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                        key
                    )
                )
            if key not in df_row.keys():
                return False
            if isinstance(df_row[key], str):
                if not isinstance(single_value, str):
                    raise InvalidQueryFilter(
                        "Value for attribute {} must be a string.".format(key)
                    )
                return re.match(single_value + r"\Z", df_row[key]) is not None
            if isinstance(df_row[key], Real):
                if isinstance(single_value, str) and single_value.lower().startswith(
                    compops
                ):
                    # compare nan metric value to numeric query
                    # (e.g. np.nan > 5)
                    if pd.isnull(df_row[key]):
                        nan_str = "np.nan"
                        # compare nan metric value to nan query
                        # (e.g., np.nan == np.nan)
                        if nan_str in single_value:
                            return eval("pd.isnull({}) == True".format(nan_str))
                        return eval("{} {}".format(nan_str, single_value))
                    elif np.isinf(df_row[key]):
                        inf_str = "np.inf"
                        # compare inf metric value to inf query
                        # (e.g., np.inf == np.inf)
                        if inf_str in single_value:
                            return eval("np.isinf({}) == True".format(inf_str))
                        return eval("{} {}".format(inf_str, single_value))
                    else:
                        return eval("{} {}".format(df_row[key], single_value))

                if isinstance(single_value, Real):
                    return df_row[key] == single_value
                raise InvalidQueryFilter(
                    "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                        key
                    )
                )
            raise InvalidQueryFilter(
                "Filter must be one of the following:\n  * A regex string for a String attribute\n  * A string starting with a comparison operator for a Numeric attribute\n  * A number for a Numeric attribute\n"
            )

        matches = True
        for k, v in attr_filter.items():
            try:
                _ = iter(v)
                # Manually raise TypeError if v is a string so that
                # the string is processed as a non-iterable
                if isinstance(v, str):
                    raise TypeError
            # Runs if v is not iterable (e.g., list, tuple, etc.)
            except TypeError:
                matches = matches and filter_single_series(df_row, k, v)
            else:
                for single_value in v:
                    matches = matches and filter_single_series(df_row, k, single_value)
        return matches

    def filter_dframe(df_row):
        if multi_index_mode == "off":
            raise MultiIndexModeMismatch(
                "The ObjectQuery's 'multi_index_mode' argument \
                cannot be set to 'off' when using multi-indexed data"
            )

        def filter_single_dframe(node, df_row, key, single_value):
            if key == "depth":
                if isinstance(single_value, str) and single_value.lower().startswith(
                    compops
                ):
                    return eval("{} {}".format(node._depth, single_value))
                if isinstance(single_value, Real):
                    return node._depth == single_value
                raise InvalidQueryFilter(
                    "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                        key
                    )
                )
            if key == "node_id":
                if isinstance(single_value, str) and single_value.lower().startswith(
                    compops
                ):
                    return eval("{} {}".format(node._hatchet_nid, single_value))
                if isinstance(single_value, Real):
                    return node._hatchet_nid == single_value
                raise InvalidQueryFilter(
                    "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                        key
                    )
                )
            if key not in df_row.columns:
                return False
            if is_string_dtype(df_row[key]):
                if not isinstance(single_value, str):
                    raise InvalidQueryFilter(
                        "Value for attribute {} must be a string.".format(key)
                    )
                apply_ret = df_row[key].apply(
                    lambda x: re.match(single_value + r"\Z", x) is not None
                )
                return _process_multi_index_mode(apply_ret, multi_index_mode)
            if is_numeric_dtype(df_row[key]):
                if isinstance(single_value, str) and single_value.lower().startswith(
                    compops
                ):
                    apply_ret = df_row[key].apply(
                        lambda x: eval("{} {}".format(x, single_value))
                    )
                    return _process_multi_index_mode(apply_ret, multi_index_mode)
                if isinstance(single_value, Real):
                    apply_ret = df_row[key].apply(lambda x: x == single_value).any()
                    return _process_multi_index_mode(apply_ret, multi_index_mode)
                raise InvalidQueryFilter(
                    "Attribute {} has a numeric type. Valid filters for this attribute are a string starting with a comparison operator or a real number.".format(
                        key
                    )
                )
            raise InvalidQueryFilter(
                "Filter must be one of the following:\n  * A regex string for a String attribute\n  * A string starting with a comparison operator for a Numeric attribute\n  * A number for a Numeric attribute\n"
            )

        matches = True
        node = df_row.name.to_frame().index[0][0]
        for k, v in attr_filter.items():
            try:
                _ = iter(v)
                if isinstance(v, str):
                    raise TypeError
            except TypeError:
                matches = matches and filter_single_dframe(node, df_row, k, v)
            else:
                for single_value in v:
                    matches = matches and filter_single_dframe(
                        node, df_row, k, single_value
                    )
        return matches

    def filter_choice(df_row):
        if isinstance(df_row, pd.DataFrame):
            return filter_dframe(df_row)
        return filter_series(df_row)

    return filter_choice if attr_filter != {} else lambda row: True


class ObjectQuery(Query):

    """Class for representing and parsing queries using the Object-based dialect."""

    def __init__(self, query, multi_index_mode="off"):
        """Builds a new ObjectQuery from an instance of the Object-based dialect syntax.

        Arguments:
            query (list): the Object-based dialect query to parse and store
        """
        if sys.version_info[0] == 2:
            super(ObjectQuery, self).__init__()
        else:
            super().__init__()
        assert isinstance(query, list)
        assert multi_index_mode in ["off", "all", "any"]
        for qnode in query:
            if isinstance(qnode, dict):
                self._add_node(predicate=_process_predicate(qnode, multi_index_mode))
            elif isinstance(qnode, str) or isinstance(qnode, int):
                self._add_node(quantifer=qnode)
            elif isinstance(qnode, tuple):
                assert isinstance(qnode[1], dict)
                if isinstance(qnode[0], str) or isinstance(qnode[0], int):
                    self._add_node(
                        qnode[0], _process_predicate(qnode[1], multi_index_mode)
                    )
                else:
                    raise InvalidQueryPath(
                        "The first value of a tuple entry in a path must be either a string or integer."
                    )
            else:
                raise InvalidQueryPath(
                    "A query path must be a list containing String, Integer, Dict, or Tuple elements"
                )



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from abc import abstractmethod

try:
    from abc import ABC
except ImportError:
    from abc import ABCMeta

    ABC = ABCMeta("ABC", (object,), {"__slots__": ()})
import sys
import warnings

from .query import Query
from .compound import (
    CompoundQuery,
    ConjunctionQuery,
    DisjunctionQuery,
    ExclusiveDisjunctionQuery,
    NegationQuery,
)
from .object_dialect import ObjectQuery
from .string_dialect import parse_string_dialect
from .engine import QueryEngine
from .errors import BadNumberNaryQueryArgs, InvalidQueryPath


# QueryEngine object for running the legacy "apply" methods
COMPATABILITY_ENGINE = QueryEngine()


class AbstractQuery(ABC):

    """Base class for all 'old-style' queries."""

    @abstractmethod
    def apply(self, gf):
        pass

    def __and__(self, other):
        """Create a new AndQuery using this query and another.

        Arguments:
            other (AbstractQuery): the other query to use in constructing the AndQuery

        Returns:
            (AndQuery): a new AndQuery that performs the AND of the results of both queries
        """
        return AndQuery(self, other)

    def __or__(self, other):
        """Create a new OrQuery using this query and another.

        Arguments:
            other (AbstractQuery): the other query to use in constructing the OrQuery

        Returns:
            (OrQuery): a new OrQuery that performs the OR of the results of both queries
        """
        return OrQuery(self, other)

    def __xor__(self, other):
        """Create a new XorQuery using this query and another.

        Arguments:
            other (AbstractQuery): the other query to use in constructing the XorQuery

        Returns:
            (XorQuery): a new XorQuery that performs the XOR of the results of both queries
        """
        return XorQuery(self, other)

    def __invert__(self):
        """Create a new NotQuery using this query.

        Returns:
            (NotQuery): a new NotQuery that inverts the results of this query
        """
        return NotQuery(self)

    @abstractmethod
    def _get_new_query(self):
        pass


class NaryQuery(AbstractQuery):

    """Base class for all compound queries that act on
    and merged N separate subqueries."""

    def __init__(self, *args):
        """Create a new NaryQuery object.

        Arguments:
            *args (AbstractQuery, str, or list): the subqueries to be performed
        """
        self.compat_subqueries = []
        if isinstance(args[0], tuple) and len(args) == 1:
            args = args[0]
        for query in args:
            if isinstance(query, list):
                self.compat_subqueries.append(QueryMatcher(query))
            elif isinstance(query, str):
                self.compat_subqueries.append(CypherQuery(query))
            elif issubclass(type(query), AbstractQuery):
                self.compat_subqueries.append(query)
            elif issubclass(type(query), Query) or issubclass(
                type(query), CompoundQuery
            ):
                self.compat_subqueries.append(query)
            else:
                raise TypeError(
                    "Subqueries for NaryQuery must be either a\
                     high-level query or a subclass of AbstractQuery"
                )

    def apply(self, gf):
        """Applies the query to the specified GraphFrame.

        Arguments:
            gf (GraphFrame): the GraphFramme on which to apply the query

        Results:
            (list): A list of nodes representing the result of the query
        """
        true_query = self._get_new_query()
        return COMPATABILITY_ENGINE.apply(true_query, gf.graph, gf.dataframe)

    def _get_new_query(self):
        """Gets all the underlying 'new-style' queries in this object.

        Returns:
            (List[Union[Query, CompoundQuery]]): a list of all the underlying 'new-style' queries in this object
        """
        true_subqueries = []
        for subq in self.compat_subqueries:
            true_subq = subq
            if issubclass(type(subq), AbstractQuery):
                true_subq = subq._get_new_query()
            true_subqueries.append(true_subq)
        return self._convert_to_new_query(true_subqueries)

    @abstractmethod
    def _convert_to_new_query(self, subqueries):
        pass


class AndQuery(NaryQuery):

    """Compound query that returns the intersection of the results
    of the subqueries."""

    def __init__(self, *args):
        """Create a new AndQuery object.

        Arguments:
            *args (AbstractQuery, str, or list): the subqueries to be performed
        """
        warnings.warn(
            "Old-style queries are deprecated and will be removed in the \
                    future. Please use new-style queries (e.g., \
                    hatchet.query.ConjunctionQuery) instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        if sys.version_info[0] == 2:
            super(AndQuery, self).__init__(*args)
        else:
            super().__init__(*args)
        if len(self.compat_subqueries) < 2:
            raise BadNumberNaryQueryArgs("AndQuery requires 2 or more subqueries")

    def _convert_to_new_query(self, subqueries):
        return ConjunctionQuery(*subqueries)


"""Alias of AndQuery"""
IntersectionQuery = AndQuery


class OrQuery(NaryQuery):

    """Compound query that returns the union of the results
    of the subqueries"""

    def __init__(self, *args):
        """Create a new OrQuery object.

        Arguments:
            *args (AbstractQuery, str, or list): the subqueries to be performed
        """
        warnings.warn(
            "Old-style queries are deprecated and will be removed in the \
                    future. Please use new-style queries (e.g., \
                    hatchet.query.DisjunctionQuery) instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        if sys.version_info[0] == 2:
            super(OrQuery, self).__init__(*args)
        else:
            super().__init__(*args)
        if len(self.compat_subqueries) < 2:
            raise BadNumberNaryQueryArgs("OrQuery requires 2 or more subqueries")

    def _convert_to_new_query(self, subqueries):
        return DisjunctionQuery(*subqueries)


"""Alias of OrQuery"""
UnionQuery = OrQuery


class XorQuery(NaryQuery):

    """Compound query that returns the symmetric difference
    (i.e., set-based XOR) of the results of the subqueries"""

    def __init__(self, *args):
        """Create a new XorQuery object.

        Arguments:
            *args (AbstractQuery, str, or list): the subqueries to be performed
        """
        warnings.warn(
            "Old-style queries are deprecated and will be removed in the \
                    future. Please use new-style queries (e.g., \
                    hatchet.query.ExclusiveDisjunctionQuery) instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        if sys.version_info[0] == 2:
            super(XorQuery, self).__init__(*args)
        else:
            super().__init__(*args)
        if len(self.compat_subqueries) < 2:
            raise BadNumberNaryQueryArgs("XorQuery requires 2 or more subqueries")

    def _convert_to_new_query(self, subqueries):
        return ExclusiveDisjunctionQuery(*subqueries)


"""Alias of XorQuery"""
SymDifferenceQuery = XorQuery


class NotQuery(NaryQuery):

    """Compound query that returns all nodes in the GraphFrame that
    are not returned from the subquery."""

    def __init__(self, *args):
        """Create a new NotQuery object.

        Arguments:
            *args (AbstractQuery, str, or list): the subquery to be performed
        """
        warnings.warn(
            "Old-style queries are deprecated and will be removed in the \
                    future. Please use new-style queries (e.g., \
                    hatchet.query.NegationQuery) instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        if sys.version_info[0] == 2:
            super(NotQuery, self).__init__(*args)
        else:
            super().__init__(*args)
        if len(self.compat_subqueries) < 1:
            raise BadNumberNaryQueryArgs("NotQuery requires exactly 1 subquery")

    def _convert_to_new_query(self, subqueries):
        return NegationQuery(*subqueries)


class QueryMatcher(AbstractQuery):

    """Processes and applies base syntax queries and Object-based queries to GraphFrames."""

    def __init__(self, query=None):
        """Create a new QueryMatcher object.

        Arguments:
            query (list, optional): if provided, convert the Object-based query
                                    into its internal representation
        """
        warnings.warn(
            "Old-style queries are deprecated and will be removed in the \
                    future. Please use new-style queries instead. \
                    For QueryMatcher, the equivalent new-style queries are \
                    hatchet.query.Query for base-syntax queries and \
                    hatchet.query.ObjectQuery for the object-dialect.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.true_query = None
        if query is None:
            self.true_query = Query()
        elif isinstance(query, list):
            self.true_query = ObjectQuery(query)
        else:
            raise InvalidQueryPath("Provided query is not a valid object dialect query")

    def match(self, wildcard_spec=".", filter_func=lambda row: True):
        """Start a query with a root node described by the arguments.

        Arguments:
            wildcard_spec (str, optional): the wildcard status of the node
            filter_func (Callable, optional): a callable acceepting only a row from a pandas DataFrame
                                              that is used to filter this node in the query

        Returns:
            (QueryMatcher): the instance of the class that called this function
        """
        self.true_query.match(wildcard_spec, filter_func)
        return self

    def rel(self, wildcard_spec=".", filter_func=lambda row: True):
        """Add another edge and node to the query.

        Arguments:
            wildcard_spec (str, optional): the wildcard status of the node
            filter_func (Callable, optional): a callable acceepting only a row from a pandas DataFrame
                                              that is used to filter this node in the query

        Returns:
            (QueryMatcher): the instance of the class that called this function
        """
        self.true_query.rel(wildcard_spec, filter_func)
        return self

    def apply(self, gf):
        """Apply the query to a GraphFrame.

        Arguments:
            gf (GraphFrame): the GraphFrame on which to apply the query

        Returns:
            (list): A list representing the set of nodes from paths that match this query
        """
        return COMPATABILITY_ENGINE.apply(self.true_query, gf.graph, gf.dataframe)

    def _get_new_query(self):
        """Get all the underlying 'new-style' query in this object.

        Returns:
            (Query or ObjectQuery): the underlying 'new-style' query in this object
        """
        return self.true_query


class CypherQuery(QueryMatcher):

    """Processes and applies Strinb-based queries to GraphFrames."""

    def __init__(self, cypher_query):
        """Create a new Cypher object.

        Arguments:
            cypher_query (str): the String-based query
        """
        warnings.warn(
            "Old-style queries are deprecated and will be removed in the \
                    future. Please use new-style queries instead. \
                    For CypherQuery, the equivalent new-style query is \
                    hatchet.query.StringQuery.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.true_query = parse_string_dialect(cypher_query)

    def _get_new_query(self):
        """Gets all the underlying 'new-style' queries in this object.

        Returns:
            (List[Union[Query, CompoundQuery]]): a list of all the underlying 'new-style' queries in this object
        """
        return self.true_query


def parse_cypher_query(cypher_query):
    """Parse all types of String-based queries, including multi-queries that
    leverage the curly brace delimiters.

    Arguments:
        cypher_query (str): the String-based query to be parsed

    Returns:
        (CypherQuery): a Hatchet query for this String-based query
    """
    warnings.warn(
        "Old-style queries are deprecated and will be removed in the \
                future. Please use new-style queries instead. \
                For parse_cypher_query, the equivalent new-style function is \
                hatchet.query.parse_string_dialect.",
        DeprecationWarning,
        stacklevel=2,
    )
    return CypherQuery(cypher_query)



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from __future__ import division

from hatchet import GraphFrame


def test_filter(mock_graph_literal):
    """Test the filter operation with a foo-bar tree."""
    gf = GraphFrame.from_literal(mock_graph_literal)

    filtered_gf = gf.filter(lambda x: x["time"] > 5.0, squash=False)
    assert len(filtered_gf.dataframe) == 9
    assert all(time > 5.0 for time in filtered_gf.dataframe["time"])

    filtered_gf = gf.filter(lambda x: x["name"].startswith("g"), squash=False)
    assert len(filtered_gf.dataframe) == 7
    assert all(name.startswith("g") for name in filtered_gf.dataframe["name"])


def test_add(mock_graph_literal):
    gf1 = GraphFrame.from_literal(mock_graph_literal)
    gf2 = GraphFrame.from_literal(mock_graph_literal)

    assert gf1.graph is not gf2.graph

    gf3 = gf1.add(gf2)

    assert gf3.graph == gf1.graph.union(gf2.graph)

    assert len(gf3.graph) == gf3.dataframe.shape[0]
    assert gf3.dataframe["time"].sum() == 330
    assert gf3.dataframe["time (inc)"].sum() == 1320

    gf4 = gf3.copy()
    assert gf4.graph is gf3.graph

    gf5 = gf3.add(gf4)
    assert gf5.graph == gf3.graph == gf4.graph


def test_sub(mock_graph_literal):
    gf1 = GraphFrame.from_literal(mock_graph_literal)
    gf2 = GraphFrame.from_literal(mock_graph_literal)

    assert gf1.graph is not gf2.graph

    gf3 = gf1.sub(gf2)

    assert gf3.graph == gf1.graph.union(gf2.graph)
    assert len(gf3.graph) == gf3.dataframe.shape[0]

    for metric in gf3.exc_metrics + gf3.inc_metrics:
        assert gf3.dataframe[metric].sum() == 0

    gf4 = gf3.copy()
    assert gf4.graph is gf3.graph

    gf5 = gf3.sub(gf4)
    assert gf5.graph == gf3.graph == gf4.graph


def test_div(mock_graph_literal):
    gf1 = GraphFrame.from_literal(mock_graph_literal)
    gf2 = GraphFrame.from_literal(mock_graph_literal)

    assert gf1.graph is not gf2.graph

    gf3 = gf1.div(gf2)
    assert len(gf3.graph) == gf3.dataframe.shape[0]

    assert gf3.graph == gf1.graph.union(gf2.graph)

    assert gf3.dataframe["time"].sum() == 21
    assert gf3.dataframe["time (inc)"].sum() == 24

    gf4 = gf3.copy()
    assert gf4.graph is gf3.graph

    gf5 = gf3.div(gf4)
    assert gf5.graph == gf3.graph == gf4.graph


def test_mul(mock_graph_literal):
    gf1 = GraphFrame.from_literal(mock_graph_literal)
    gf2 = GraphFrame.from_literal(mock_graph_literal)

    assert gf1.graph is not gf2.graph

    gf3 = gf1.mul(gf2)
    assert len(gf3.graph) == gf3.dataframe.shape[0]

    assert gf3.graph == gf1.graph.union(gf2.graph)

    assert gf3.dataframe["time"].sum() == 1575
    assert gf3.dataframe["time (inc)"].sum() == 37900


def test_add_operator(mock_graph_literal):
    gf1 = GraphFrame.from_literal(mock_graph_literal)
    gf2 = GraphFrame.from_literal(mock_graph_literal)

    assert gf1.graph is not gf2.graph

    gf3 = gf1 + gf2

    assert gf3.graph == gf1.graph.union(gf2.graph)
    assert len(gf3.graph) == gf3.dataframe.shape[0]

    assert gf3.dataframe["time"].sum() == 330
    assert gf3.dataframe["time (inc)"].sum() == 1320

    gf4 = gf3.copy()
    assert gf4.graph is gf3.graph

    gf5 = gf3 + gf4
    assert gf5.graph == gf3.graph == gf4.graph

    gf6 = gf1 + gf2 + gf1
    assert gf6.dataframe["time"].sum() == 495

    gf7 = gf1 + gf2
    gf8 = gf7 + gf1
    assert gf8.graph == gf6.graph
    assert gf8.dataframe["time"].sum() == gf6.dataframe["time"].sum()


def test_sub_operator(mock_graph_literal):
    gf1 = GraphFrame.from_literal(mock_graph_literal)
    gf2 = GraphFrame.from_literal(mock_graph_literal)

    assert gf1.graph is not gf2.graph

    gf3 = gf1 - gf2

    assert gf3.graph == gf1.graph.union(gf2.graph)
    assert len(gf3.graph) == gf3.dataframe.shape[0]

    for metric in gf3.exc_metrics + gf3.inc_metrics:
        assert gf3.dataframe[metric].sum() == 0

    gf4 = gf3.copy()
    assert gf4.graph is gf3.graph

    gf5 = gf3.sub(gf4)

    assert gf5.graph == gf3.graph == gf4.graph

    gf6 = gf1 - gf2 - gf1
    assert gf6.dataframe["time"].sum() == -165

    gf7 = gf1 - gf2
    gf8 = gf7 - gf1
    assert gf8.graph == gf6.graph
    assert gf8.dataframe["time"].sum() == gf6.dataframe["time"].sum()


def test_div_operator(mock_graph_literal):
    gf1 = GraphFrame.from_literal(mock_graph_literal)
    gf2 = GraphFrame.from_literal(mock_graph_literal)

    assert gf1.graph is not gf2.graph

    gf3 = gf1 / gf2

    assert gf3.graph == gf1.graph.union(gf2.graph)
    assert len(gf3.graph) == gf3.dataframe.shape[0]

    assert gf3.dataframe["time"].sum() == 21
    assert gf3.dataframe["time (inc)"].sum() == 24

    gf4 = gf3.copy()
    assert gf4.graph is gf3.graph

    gf5 = gf3 / gf4 / gf3

    assert gf5.graph == gf3.graph == gf4.graph
    assert gf5.dataframe["time (inc)"].sum() == 24

    gf6 = gf3 / gf4
    gf7 = gf6 / gf3
    assert gf7.graph == gf5.graph
    assert gf7.dataframe["time"].sum() == gf5.dataframe["time"].sum()


def test_mul_operator(mock_graph_literal):
    gf1 = GraphFrame.from_literal(mock_graph_literal)
    gf2 = GraphFrame.from_literal(mock_graph_literal)
    gf3 = GraphFrame.from_literal(mock_graph_literal)

    assert gf1.graph is not gf2.graph is not gf3.graph

    gf4 = gf1 * gf2 * gf3

    assert gf4.graph == gf1.graph.union(gf2.graph.union(gf3.graph))
    assert len(gf4.graph) == gf4.dataframe.shape[0]

    assert gf4.dataframe["time"].sum() == 17625
    assert gf4.dataframe["time (inc)"].sum() == 3397500


def test_iadd_operator(mock_graph_literal):
    gf1 = GraphFrame.from_literal(mock_graph_literal)
    gf2 = GraphFrame.from_literal(mock_graph_literal)

    assert gf1.graph is not gf2.graph

    gf1 += gf2

    assert gf1.graph == gf1.graph.union(gf2.graph)
    assert len(gf1.graph) == gf1.dataframe.shape[0]

    assert gf1.dataframe["time"].sum() == 330
    assert gf1.dataframe["time (inc)"].sum() == 1320

    gf3 = gf1.copy()
    assert gf3.graph is gf1.graph

    gf3 += gf1 + gf2 + gf2

    assert gf3.graph == gf1.graph
    assert gf3.dataframe["time"].sum() == 990


def test_isub_operator(mock_graph_literal):
    gf1 = GraphFrame.from_literal(mock_graph_literal)
    gf2 = GraphFrame.from_literal(mock_graph_literal)

    assert gf1.graph is not gf2.graph

    gf1 -= gf2

    assert gf1.graph == gf1.graph.union(gf2.graph)
    assert len(gf1.graph) == gf1.dataframe.shape[0]

    for metric in gf1.exc_metrics + gf1.inc_metrics:
        assert gf1.dataframe[metric].sum() == 0

    gf3 = gf1.copy()
    assert gf3.graph is gf1.graph

    gf3 -= gf1

    assert gf3.graph == gf1.graph


def test_idiv_operator(mock_graph_literal):
    gf1 = GraphFrame.from_literal(mock_graph_literal)
    gf2 = GraphFrame.from_literal(mock_graph_literal)

    assert gf1.graph is not gf2.graph

    gf1 /= gf2

    assert gf1.graph == gf1.graph.union(gf2.graph)
    assert len(gf1.graph) == gf1.dataframe.shape[0]

    assert gf1.dataframe["time"].sum() == 21
    assert gf1.dataframe["time (inc)"].sum() == 24

    gf3 = gf1.copy()
    assert gf3.graph is gf1.graph

    gf3 /= gf1

    assert gf3.graph == gf1.graph


def test_imul_operator(mock_graph_literal):
    gf1 = GraphFrame.from_literal(mock_graph_literal)
    gf2 = GraphFrame.from_literal(mock_graph_literal)

    assert gf1.graph is not gf2.graph

    gf1 *= gf2

    assert gf1.graph == gf1.graph.union(gf2.graph)
    assert len(gf1.graph) == gf1.dataframe.shape[0]

    assert gf1.dataframe["time"].sum() == 1575
    assert gf1.dataframe["time (inc)"].sum() == 37900



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import numpy as np

from hatchet import GraphFrame
from hatchet.readers.gprof_dot_reader import GprofDotReader

roots = [
    "psm_no_lock",
    "(below main)",
    "0x0000000000000ff0",
    "_dl_catch_error'2",
    "hwloc_topology_load",
    "ibv_get_device_list",
    "psm_allocate_vbufs",
]


def test_graphframe(calc_pi_callgrind_dot):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_gprof_dot(str(calc_pi_callgrind_dot))

    assert len(gf.dataframe.groupby("name")) == 105

    for col in gf.dataframe.columns:
        if col in ("time (inc)", "time"):
            assert gf.dataframe[col].dtype == np.float64
        elif col in ("name", "module", "node"):
            assert gf.dataframe[col].dtype == object

    # TODO: add tests to confirm values in dataframe


def test_read_calc_pi_database(calc_pi_callgrind_dot):
    """Sanity check the GprofDot reader by examining a known input."""
    reader = GprofDotReader(str(calc_pi_callgrind_dot))

    list_roots = reader.create_graph()
    root_names = []
    for root in list_roots:
        root_names.append(root.frame.attrs["name"])

    assert all(rt in root_names for rt in roots)



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from setuptools import setup
from setuptools import Extension
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# Get the version in a safe way which does not refrence hatchet `__init__` file
# per python docs: https://packaging.python.org/guides/single-sourcing-package-version/
version = {}
with open("./hatchet/version.py") as fp:
    exec(fp.read(), version)


cmd_class = {}
ext_modules = []
should_cythonize = False
mod_import_path = "hatchet.cython_modules.libs"
mod_file_path = "hatchet/cython_modules"
mod_names = [
    "reader_modules",
    "graphframe_modules",
]


for mname in mod_names:
    c_file = path.join(mod_file_path, "{}.c".format(mname))
    pyx_file = path.join(mod_file_path, "{}.pyx".format(mname))
    if not path.isfile(pyx_file):
        raise FileNotFoundError(
            "Requested Cython extension not found: {}".format(pyx_file)
        )
    if path.isfile(c_file):
        should_cythonize = False
        ext_modules.append(
            Extension(
                "{}.{}".format(mod_import_path, mname),
                [c_file],
            )
        )
    else:
        should_cythonize = True
        ext_modules.append(
            Extension(
                "{}.{}".format(mod_import_path, mname),
                [pyx_file],
            )
        )

if should_cythonize:
    from Cython.Build import cythonize, build_ext

    ext_modules = cythonize(ext_modules)
    cmd_class.update({"build_ext": build_ext})


ext_modules = [
    Extension(
        "hatchet.cython_modules.libs.reader_modules",
        ["hatchet/cython_modules/reader_modules.pyx"],
    ),
    Extension(
        "hatchet.cython_modules.libs.graphframe_modules",
        ["hatchet/cython_modules/graphframe_modules.pyx"],
    ),
]


setup(
    name="llnl-hatchet",
    version=version["__version__"],
    license="MIT",
    description="A Python library for analyzing hierarchical performance data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/llnl/hatchet",
    project_urls={
        "Source Code": "https://github.com/llnl/hatchet",
        "Documentation": "https://llnl-hatchet.readthedocs.io/en/latest/",
    },
    author="Stephanie Brink",
    author_email="brink2@llnl.gov",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
    ],
    keywords="",
    python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*",
    packages=[
        "hatchet",
        "hatchet.readers",
        "hatchet.writers",
        "hatchet.query",
        "hatchet.vis",
        "hatchet.util",
        "hatchet.external",
        "hatchet.tests",
        "hatchet.cython_modules.libs",
    ],
    include_package_data=True,
    install_requires=[
        "pydot",
        "PyYAML",
        "matplotlib",
        "numpy",
        "pandas",
        "textX < 3.0.0; python_version < '3.6'",
        "textX >= 3.0.0; python_version >= '3.6'",
        "multiprocess",
        "caliper-reader",
    ],
    ext_modules=ext_modules,
    cmdclass=cmd_class,
)



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from abc import abstractmethod

import sys

from .query import Query
from .string_dialect import parse_string_dialect
from .object_dialect import ObjectQuery
from .errors import BadNumberNaryQueryArgs


class CompoundQuery(object):

    """Base class for all types of compound queries."""

    def __init__(self, *queries):
        """Collect the provided queries into a list, constructing ObjectQuery and StringQuery objects as needed.

        Arguments:
            *queries (Query, CompoundQuery, list, or str): the subqueries of the compound query
        """
        self.subqueries = []
        if isinstance(queries[0], tuple) and len(queries) == 1:
            queries = queries[0]
        for query in queries:
            if issubclass(type(query), Query) or issubclass(type(query), CompoundQuery):
                self.subqueries.append(query)
            elif isinstance(query, list):
                self.subqueries.append(ObjectQuery(query))
            elif isinstance(query, str):
                self.subqueries.append(parse_string_dialect(query))
            else:
                raise TypeError(
                    "Subqueries for NaryQuery must be either a \
                     high-level query or a subclass of AbstractQuery"
                )

    @abstractmethod
    def _apply_op_to_results(self, subquery_results):
        """Combines/Modifies the results of the subqueries based on the operation the subclass
        represents.
        """
        pass


class ConjunctionQuery(CompoundQuery):

    """A compound query that combines the results of its subqueries
    using set conjunction.
    """

    def __init__(self, *queries):
        """Create the ConjunctionQuery.

        Arguments:
            *queries (Query, CompoundQuery, list, or str): the subqueries of the compound query
        """
        if sys.version_info[0] == 2:
            super(ConjunctionQuery, self).__init__(*queries)
        else:
            super().__init__(*queries)
        if len(self.subqueries) < 2:
            raise BadNumberNaryQueryArgs(
                "ConjunctionQuery requires 2 or more subqueries"
            )

    def _apply_op_to_results(self, subquery_results, graph):
        """Combines the results of the subqueries using set conjunction.

        Arguments:
            subquery_results (list): a list containing the results of each subquery
            graph (hatchet.Graph): the graph associated with the data being queried

        Returns:
            (list): A list containing all the nodes satisfying the conjunction of the subqueries' results
        """
        intersection_set = set(subquery_results[0]).intersection(*subquery_results[1:])
        return list(intersection_set)


class DisjunctionQuery(CompoundQuery):

    """A compound query that combines the results of its subqueries
    using set disjunction.
    """

    def __init__(self, *queries):
        """Create the DisjunctionQuery.

        Arguments:
            *queries (Query, CompoundQuery, list, or str): the subqueries of the compound query
        """
        if sys.version_info[0] == 2:
            super(DisjunctionQuery, self).__init__(*queries)
        else:
            super().__init__(*queries)
        if len(self.subqueries) < 2:
            raise BadNumberNaryQueryArgs(
                "DisjunctionQuery requires 2 or more subqueries"
            )

    def _apply_op_to_results(self, subquery_results, graph):
        """Combines the results of the subqueries using set disjunction.

        Arguments:
            subquery_results (list): a list containing the results of each subquery
            graph (hatchet.Graph): the graph associated with the data being queried

        Returns:
            (list): A list containing all the nodes satisfying the disjunction of the subqueries' results
        """
        union_set = set().union(*subquery_results)
        return list(union_set)


class ExclusiveDisjunctionQuery(CompoundQuery):

    """A compound query that combines the results of its subqueries
    using exclusive set disjunction.
    """

    def __init__(self, *queries):
        """Create the ExclusiveDisjunctionQuery.

        Arguments:
            *queries (Query, CompoundQuery, list, or str): the subqueries of the compound query
        """
        if sys.version_info[0] == 2:
            super(ExclusiveDisjunctionQuery, self).__init__(*queries)
        else:
            super().__init__(*queries)
        if len(self.subqueries) < 2:
            raise BadNumberNaryQueryArgs("XorQuery requires 2 or more subqueries")

    def _apply_op_to_results(self, subquery_results, graph):
        """Combines the results of the subqueries using exclusive set disjunction.

        Arguments:
            subquery_results (list): a list containing the results of each subquery
            graph (hatchet.Graph): the graph associated with the data being queried

        Returns:
            (list): A list containing all the nodes satisfying the exclusive disjunction of the subqueries' results
        """
        xor_set = set()
        for res in subquery_results:
            xor_set = xor_set.symmetric_difference(set(res))
        return list(xor_set)


class NegationQuery(CompoundQuery):

    """A compound query that inverts/negates the result of
    its single subquery.
    """

    def __init__(self, *queries):
        """Create the NegationQuery.

        Arguments:
            *queries (Query, CompoundQuery, list, or str): the subqueries of the compound query. There must be eactly one subquery provided.
        """
        if sys.version_info[0] == 2:
            super(NegationQuery, self).__init__(*queries)
        else:
            super().__init__(*queries)
        if len(self.subqueries) != 1:
            raise BadNumberNaryQueryArgs("NotQuery requires exactly 1 subquery")

    def _apply_op_to_results(self, subquery_results, graph):
        """Inverts the results of the subquery so that all nodes not in the results are returned.

        Arguments:
            subquery_results (list): a list containing the results of each subquery
            graph (hatchet.Graph): the graph associated with the data being queried

        Returns:
            (list): A list containing all the nodes in the Graph not contained in the subquery's results
        """
        nodes = set(graph.traverse())
        query_nodes = set(subquery_results[0])
        return list(nodes.difference(query_nodes))



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import warnings
import pandas as pd
import sys
from .dataframe_reader import DataframeReader


class HDF5Reader(DataframeReader):
    def __init__(self, filename):
        # TODO Remove Arguments when Python 2.7 support is dropped
        if sys.version_info[0] == 2:
            super(HDF5Reader, self).__init__(filename)
        else:
            super().__init__(filename)

    def _read_dataframe_from_file(self, **kwargs):
        df = None
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=Warning)
            df = pd.read_hdf(self.filename, **kwargs)
        return df



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import json

import pandas as pd

import hatchet.graphframe
from hatchet.node import Node
from hatchet.graph import Graph
from hatchet.frame import Frame


class JsonReader:
    """Create a GraphFrame from a json string of the following format.

    Return:
        (GraphFrame): graphframe containing data from dictionaries
    """

    def __init__(self, json_spec):
        """Read from a json string specification of a graphframe

        json (string): Json specification of a graphframe.
        """
        self.spec_dict = json.loads(json_spec)

    def read(self):
        roots = []
        for graph_spec in self.spec_dict["graph"]:
            # turn frames into nodes
            for nid, value in graph_spec.items():
                graph_spec[nid]["data"] = Node(Frame(value["data"]), hnid=int(nid))

            # connect nodes
            for nid, value in graph_spec.items():
                for child in value["children"]:
                    child = str(child)
                    value["data"].add_child(graph_spec[child]["data"])
                    graph_spec[child]["data"].add_parent(value["data"])

            for nid, value in graph_spec.items():
                if len(value["data"].parents) == 0:
                    roots.append(value["data"])

        grph = Graph(roots)

        # make the dataframes
        dataframe = pd.DataFrame(self.spec_dict["dataframe"])
        for graph_spec in self.spec_dict["graph"]:
            dataframe["node"] = dataframe["node"].map(
                lambda n: graph_spec[str(n)]["data"] if (str(n) in graph_spec) else n
            )
        dataframe.set_index(self.spec_dict["dataframe_indices"], inplace=True)

        return hatchet.graphframe.GraphFrame(
            grph,
            dataframe,
            self.spec_dict["exclusive_metrics"],
            self.spec_dict["inclusive_metrics"],
        )



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT


class InvalidQueryPath(Exception):
    """Raised when a query does not have the correct syntax"""


class InvalidQueryFilter(Exception):
    """Raised when a query filter does not have a valid syntax"""


class RedundantQueryFilterWarning(Warning):
    """Warned when a query filter does nothing or is redundant"""


class BadNumberNaryQueryArgs(Exception):
    """Raised when a query filter does not have a valid syntax"""


class MultiIndexModeMismatch(Exception):
    """Raised when an ObjectQuery or StringQuery object
    is set to use multi-indexed data, but no multi-indexed
    data is provided"""



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import numpy as np
import re

from hatchet import GraphFrame
from hatchet.external.console import ConsoleRenderer


def test_graphframe(hatchet_cycle_pstats):
    gf = GraphFrame.from_cprofile(str(hatchet_cycle_pstats))

    assert len(gf.dataframe.groupby("file")) == 4
    assert len(gf.dataframe.groupby("name")) == 9

    gf.dataframe.reset_index(inplace=True)

    for col in gf.dataframe.columns:
        if col in ("time (inc)", "time"):
            assert gf.dataframe[col].dtype == np.float64
        elif col in ("line", "numcalls", "nativecalls"):
            assert gf.dataframe[col].dtype == np.int64
        elif col in ("name", "type", "file", "module", "node"):
            assert gf.dataframe[col].dtype == object


def test_tree(hatchet_cycle_pstats):
    gf = GraphFrame.from_cprofile(str(hatchet_cycle_pstats))

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
        render_header=True,
    )
    assert "g pstats_reader_test.py" in output
    assert "<method 'disable' ...Profiler' objects> ~" in output

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time (inc)",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
        render_header=True,
    )
    assert "f pstats_reader_test.py" in output
    assert re.match("(.|\n)*recursive(.|\n)*recursive", output)
